from typing import Any, Dict, Iterable, List, Union

import onnx
from onnx import helper, mapping, numpy_helper
from xslim.defs import GLOBAL_FUNCTIONS_MAPPING, MIN_ONNX_OPSET_VERSION

from ..core import DEFAULT_OPSET_DOMAIN, GRAPH_OPSET_ATTRIB, NetworkFramework, is_file_exist
from ..IR import BaseGraph, GraphBuilder, Operation, Opset, Variable


class OnnxParser(GraphBuilder):
    def build_variables(
        self,
        graph: BaseGraph,
        graph_inputs: List[str],
        graph_outputs: List[str],
        op_inputs: Dict[str, list],
        op_outputs: Dict[str, list],
    ) -> BaseGraph:
        var_list = []

        for op_name, _ in graph.operations.items():
            for var_name in op_inputs[op_name]:
                var_list.append(var_name)
            for var_name in op_outputs[op_name]:
                var_list.append(var_name)

        # create all variable at once.
        for var_name in set(var_list):
            graph.variables[var_name] = Variable(name=var_name)

        # build graph's input, output variables.
        try:
            for var_name in graph_inputs:
                if var_name not in graph.variables:
                    continue
                graph.inputs[var_name] = graph.variables[var_name]
            for var_name in graph_outputs:
                graph.outputs[var_name] = graph.variables[var_name]
        except KeyError as e:
            raise KeyError("seems you got an input/output variable that is not linked to any operation.")

        # build operation inputs, outputs variables.
        for op in graph.operations.values():
            for var_name in op_inputs[op.name]:
                var = graph.variables[var_name]
                var.dest_ops.append(op)
                op.inputs.append(graph.variables[var_name])
            for var_name in op_outputs[op.name]:
                var = graph.variables[var_name]
                var.source_op = op
                op.outputs.append(graph.variables[var_name])
        return graph

    def initialize_params(self, graph: BaseGraph, initializer: Dict[str, Any]) -> BaseGraph:
        for var in graph.variables.values():
            if var.name in initializer:
                for dest_op in var.dest_ops:
                    assert isinstance(dest_op, Operation)
                    dest_op.parameters.append(var)
                var.value = initializer[var.name]
                var.is_parameter = True
        return graph

    def de_inplace(self, graph: BaseGraph) -> BaseGraph:
        """Remove inplace layer in netdef If the names of bottom and top are same,
        it means the computation of this layer is in place."""

        def new_name(_name):
            if _name == "":
                return ""
            elif _name not in total_write_times or current_write_times:
                return _name
            elif current_write_times[_name] == total_write_times[_name]:
                return _name
            else:
                return f"{_name}_ver{current_write_times[_name]}"

        total_write_times = {}
        for op in graph.operations.values():
            for top in op.outputs:
                total_write_times.setdefault(top._name, 0)
                total_write_times[top._name] += 1

        current_write_times = {}
        for name in graph.inputs.keys():
            total_write_times[name] = 0
            current_write_times[name] = 0

        for op in graph.operations.values():
            for bottom in op.inputs:
                if bottom.is_parameter:
                    continue
                bottom._name = new_name(bottom._name)
            for top in op.outputs:
                current_write_times.setdefault(top._name, 0)
                current_write_times[top._name] += 1
                top._name = new_name(top._name)

    def refine_graph(self, graph: BaseGraph) -> BaseGraph:
        for op in graph.operations.values():
            for key, value in op.attributes.items():
                if isinstance(value, bytes):
                    # Change bytes to string
                    value = value.decode("utf-8")
                if op.type == "Constant" or op.type == "ConstantOfShape":
                    # The attribute of 'Constant' node is a value, needs to convert to numpy array
                    value = numpy_helper.to_array(value).copy()
                if op.type == "Cast":
                    # The attribute of 'Cast' node is data type (represented in int), need to convert to numpy data type
                    value = mapping.TENSOR_TYPE_TO_NP_TYPE[value]
                op.attributes[key] = value

        graph_initializers = []
        for input_var in graph.inputs.values():
            # remove initilizer from graph.inputs
            if input_var.value is not None:
                graph_initializers.append(input_var.name)
        for non_input_var in graph_initializers:
            graph.inputs.pop(non_input_var)
        return graph

    def convert_opsets_to_str(self, opsets: Iterable) -> List[Dict[str, str]]:
        results = []
        for opset in opsets:
            results.append({"domain": opset.domain, "version": opset.version})
        return results

    def build_graph(self, graph_or_function: [onnx.GraphProto, onnx.FunctionProto], import_opset_dict) -> BaseGraph:
        _rand_seed = 0
        graph = BaseGraph(name=graph_or_function.name, built_from=NetworkFramework.ONNX)
        op_inputs_dict, op_outputs_dict = {}, {}
        for node in graph_or_function.node:
            op_name = node.name
            if len(op_name) == 0:  # some operation do not have a name, we just generate one.
                op_name = "generated_name_" + str(_rand_seed)
                _rand_seed += 1

            if op_name in graph.operations:
                raise KeyError(f"Duplicated operation {op_name} was found.")

            opset_tmp = Opset(
                domain=DEFAULT_OPSET_DOMAIN if node.domain == "" else node.domain,
                version=import_opset_dict[node.domain],
            )

            graph.operations[op_name] = Operation(
                name=op_name,
                op_type=node.op_type,
                attributes={item.name: helper.get_attribute_value(item) for item in node.attribute},
                opset=opset_tmp,
            )
            op_inputs_dict[op_name] = [var_name for var_name in node.input]
            op_outputs_dict[op_name] = [var_name for var_name in node.output]

        initializer = {}
        if isinstance(graph_or_function, onnx.GraphProto):
            for item in graph_or_function.initializer:
                init_name = item.name
                value = numpy_helper.to_array(item)
                initializer[init_name] = value

            inputs = [item.name for item in graph_or_function.input]
            outputs = [item.name for item in graph_or_function.output]
        else:
            inputs = [item for item in graph_or_function.input]
            outputs = [item for item in graph_or_function.output]
            graph._detail["function_input"] = graph_or_function.input
            graph._detail["function_output"] = graph_or_function.output
            graph._detail["function_domain"] = graph_or_function.domain
            graph._detail["function_opset_import"] = self.convert_opsets_to_str(graph_or_function.opset_import)
            graph._detail["function_attribute"] = graph_or_function.attribute
            graph._detail["function_attribute_proto"] = {
                item.name: helper.get_attribute_value(item) for item in graph_or_function.attribute_proto
            }

        graph._num_of_generated_op = len(graph.operations)
        graph._num_of_generated_var = len(graph.variables)
        graph = self.build_variables(
            graph, graph_inputs=inputs, graph_outputs=outputs, op_inputs=op_inputs_dict, op_outputs=op_outputs_dict
        )
        graph = self.initialize_params(graph, initializer)
        self.de_inplace(graph)
        self.refine_graph(graph)

        return graph

    def build(self, file_path_or_proto: Union[str, onnx.ModelProto]) -> BaseGraph:
        if isinstance(file_path_or_proto, str):
            if not is_file_exist(file_path_or_proto):
                raise FileNotFoundError(f"file {file_path_or_proto} does not exist, or it is a directory.")
            model_pb = onnx.load(file_path_or_proto)
        elif isinstance(file_path_or_proto, onnx.ModelProto):
            model_pb = file_path_or_proto
        else:
            raise TypeError("type for file_path_or_proto {} error".format(type(file_path_or_proto)))

        opsets = model_pb.opset_import

        assert isinstance(
            model_pb, onnx.ModelProto
        ), f"onnx load failed, only ProtoBuffer object is expected here, while {type(model_pb)} is loaded."
        graph_pb = model_pb.graph

        # graph = BaseGraph(name=graph_pb.name, built_from=NetworkFramework.ONNX)
        # graph._detail[GRAPH_OPSET_ATTRIB] = self.convert_opsets_to_str(opsets)

        model_opset_import = self.convert_opsets_to_str(opsets)
        import_opset_dict = {}
        for opset in model_opset_import:
            import_opset_dict[opset["domain"]] = opset["version"]

        graph = self.build_graph(graph_pb, import_opset_dict)

        graph._detail[GLOBAL_FUNCTIONS_MAPPING] = {}
        for function_proto in model_pb.functions:
            function_impl = self.build_graph(function_proto, import_opset_dict)
            graph._detail[GLOBAL_FUNCTIONS_MAPPING][
                "{}.{}".format(function_proto.domain, function_proto.name)
            ] = function_impl

        graph._detail["pb_opset_import"] = model_opset_import
        graph._detail["pb_input"] = [item for item in graph_pb.input if item.name in graph.inputs]
        graph._detail["pb_output"] = [item for item in graph_pb.output if item.name in graph.outputs]
        graph._detail["pb_functions"] = model_pb.functions
        graph._detail["pb_metadata_props"] = model_pb.metadata_props
        graph._detail["pb_model_version"] = model_pb.model_version
        graph._detail["pb_producer_name"] = model_pb.producer_name
        graph._detail["pb_producer_version"] = model_pb.producer_version
        graph._detail["pb_doc_string"] = model_pb.doc_string
        graph._detail["pb_ir_version"] = model_pb.ir_version
        return graph



"""Regression tests for FP16 ONNX conversion."""

import os
import sys
import unittest
from unittest import mock

import onnx
import onnx_graphsurgeon as osg
from onnx import TensorProto, helper
from onnxconverter_common import float16 as convert_float_to_float16

sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

from xslim.quantizer import convert_to_fp16 as convert_to_fp16_module


class TestConvertToFp16(unittest.TestCase):
    """Test dtype preservation around FP16 conversion legalization."""

    @staticmethod
    def _build_add_cast_model():
        x = helper.make_tensor_value_info("x", TensorProto.FLOAT, [1])
        y = helper.make_tensor_value_info("y", TensorProto.FLOAT, [1])
        z = helper.make_tensor_value_info("z", TensorProto.FLOAT, [1])
        bias = helper.make_tensor("bias", TensorProto.FLOAT, [1], [1.0])

        nodes = [
            helper.make_node("Add", ["x", "bias"], ["z"], name="add"),
            helper.make_node("Cast", ["z"], ["y"], name="cast", to=TensorProto.FLOAT),
        ]
        graph = helper.make_graph(nodes, "add_cast_graph", [x], [y], [bias], value_info=[z])
        model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 13)])
        return onnx.shape_inference.infer_shapes(model)

    def test_convert_to_fp16_restores_tensor_dtypes_before_legalize(self):
        model = self._build_add_cast_model()
        converted_model = convert_float_to_float16.convert_float_to_float16(
            model, keep_io_types=True, disable_shape_infer=False
        )
        converted_model.graph.ClearField("value_info")

        with mock.patch.object(
            convert_to_fp16_module.convert_float_to_float16,
            "convert_float_to_float16",
            return_value=converted_model,
        ):
            fp16_model = convert_to_fp16_module.convert_to_fp16_onnx_model(model, [], [], sim_en=False)

        osg_graph = osg.import_onnx(fp16_model)
        cast_nodes = [node for node in osg_graph.nodes if node.op == "Cast" and node.name == "cast"]

        self.assertEqual(len(cast_nodes), 1)
        self.assertEqual(cast_nodes[0].attrs["to"], TensorProto.FLOAT)
        self.assertEqual(cast_nodes[0].inputs[0].dtype, onnx.helper.tensor_dtype_to_np_dtype(TensorProto.FLOAT16))
        self.assertEqual(cast_nodes[0].outputs[0].dtype, onnx.helper.tensor_dtype_to_np_dtype(TensorProto.FLOAT))
        self.assertGreater(len(fp16_model.graph.value_info), 0)


if __name__ == "__main__":
    unittest.main()



import torch
from numpy import dot, ndarray
from numpy.linalg import norm


def torch_cosine_similarity(y_pred: torch.Tensor, y_real: torch.Tensor, reduction: str = "mean") -> torch.Tensor:
    if y_pred.shape != y_real.shape:
        raise ValueError(
            f"Can not compute mse loss for tensors with different shape. " f"({y_pred.shape} and {y_real.shape})"
        )
    reduction = str(reduction).lower()

    while y_pred.ndim < 2:
        y_pred = y_pred.unsqueeze(0)
        y_real = y_real.unsqueeze(0)

    y_pred = y_pred.flatten(start_dim=1).float()
    y_real = y_real.flatten(start_dim=1).float()

    cosine_sim = torch.cosine_similarity(y_pred, y_real, dim=-1)

    if reduction == "mean":
        return torch.mean(cosine_sim)
    elif reduction == "sum":
        return torch.sum(cosine_sim)
    elif reduction == "none":
        return cosine_sim
    else:
        raise ValueError(f"Unsupported reduction method.")


def numpy_cosine_similarity(x: ndarray, y: ndarray) -> ndarray:
    return dot(x, y) / (norm(x) * norm(y))


def torch_cosine_similarity_as_loss(
    y_pred: torch.Tensor, y_real: torch.Tensor, reduction: str = "mean"
) -> torch.Tensor:
    return 1 - torch_cosine_similarity(y_pred=y_pred, y_real=y_real, reduction=reduction)



import sys

from loguru import logger

logger.remove()
logger.add(
    sink=sys.stdout,
    format="{time:YYYY-MM-DD HH:mm:ss} | <lvl>{name:24}</> | <lvl>{level:8}</> | <lvl>{message}</>",
    colorize=True,
    level="INFO",
)


def xslim_error(info: str):
    logger.error(info)


def xslim_warning(info: str):
    logger.warning(info)


def xslim_info(info: str):
    logger.info(info)


def xslim_debug(info: str):
    logger.debug(info)


def xslim_trace(info: str):
    logger.trace(info)



from functools import partial
from typing import Dict
import math
import torch
from ....core.defs import ppq_warning
from ...measure.cosine import torch_cosine_similarity
from ...measure.norm import torch_mean_square_error, torch_snr_error


class MeasureRecorder:
    """Helper class for collecting data."""

    def __init__(self, measurement: str = "cosine", reduce: str = "mean") -> None:
        self.num_of_elements = 0
        self.measure = 0
        if reduce not in {"mean", "max"}:
            raise ValueError(f"PPQ MeasureRecorder Only support reduce by mean or max, however {reduce} was given.")

        if str(measurement).lower() == "cosine":
            measure_fn = partial(torch_cosine_similarity, reduction=reduce)
        elif str(measurement).lower() == "mse":
            measure_fn = partial(torch_mean_square_error, reduction=reduce)
        elif str(measurement).lower() == "snr":
            measure_fn = partial(torch_snr_error, reduction=reduce)
        else:
            raise ValueError(
                "Unsupported measurement detected. "
                f"PPQ only support mse, snr and consine now, while {measurement} was given."
            )

        self.measure_fn = measure_fn
        self.reduce = reduce

    def update(self, y_pred: torch.Tensor, y_real: torch.Tensor):
        elements = y_pred.shape[0]
        if elements != y_real.shape[0]:
            raise Exception(
                "Can not update measurement, cause your input data do not share a same batchsize. "
                f"Shape of y_pred {y_pred.shape} - against shape of y_real {y_real.shape}"
            )
        result = self.measure_fn(y_pred=y_pred, y_real=y_real).item()

        if self.reduce == "mean":
            self.measure = self.measure * self.num_of_elements + result * elements
            self.num_of_elements += elements
            self.measure /= self.num_of_elements

        if self.reduce == "max":
            self.measure = max(self.measure, result)
            self.num_of_elements += elements


class MeasurePrinter:
    """Helper class for print top-k record."""

    def __init__(
        self,
        data: Dict[str, float],
        measure: str,
        label: str = "Layer",
        k: int = None,
        order: str = "large_to_small",
        percentage: bool = False,
    ) -> None:
        if order not in {"large_to_small", "small_to_large", None}:
            raise ValueError('Parameter "order" can only be "large_to_small" or "small_to_large"')
        self.collection = [(name, value) for name, value in data.items()]
        if order is not None:
            self.collection = sorted(self.collection, key=lambda x: x[1])
            if order == "large_to_small":
                self.collection = self.collection[::-1]
        if k is not None:
            self.collection = self.collection[:k]

        if order is None:
            sorted_collection = sorted(self.collection, key=lambda x: x[1])
            largest_element, smallest_element = sorted_collection[-1][1], sorted_collection[0][1]
        elif order == "large_to_small":
            largest_element, smallest_element = self.collection[0][1], self.collection[-1][1]
        else:
            largest_element, smallest_element = self.collection[-1][1], self.collection[0][1]
        self.normalized_by = largest_element - smallest_element
        self.min = smallest_element

        max_name_length = len(label)
        for name, _ in self.collection:
            max_name_length = max(len(name), max_name_length)
        self.max_name_length = max_name_length
        self.measure_str = measure
        self.label = label
        self.percentage = percentage

    def print(self, max_blocks: int = 20):
        print(f'{self.label}{" " * (self.max_name_length - len(self.label))}  | {self.measure_str} ')
        for name, value in self.collection:
            normalized_value = (value - self.min) / (self.normalized_by + 1e-7)
            if math.isnan(value):
                ppq_warning("MeasurePrinter found an NaN value in your data.")
                normalized_value = 0
            num_of_blocks = round(normalized_value * max_blocks)

            if not self.percentage:
                print(
                    f'{name}:{" " * (self.max_name_length - len(name))} | '
                    f'{"█" * num_of_blocks}{" " * (max_blocks - num_of_blocks)} | '
                    f"{value:.4f}"
                )
            else:
                print(
                    f'{name}:{" " * (self.max_name_length - len(name))} | '
                    f'{"█" * num_of_blocks}{" " * (max_blocks - num_of_blocks)} | '
                    f"{value * 100:.3f}%"
                )



"""PPQ Core Foreign Function Interface PPQ 核心编程语言接口.

You are not allowed to modify this 请勿修改此文件
"""

import os
from typing import List

import torch
from torch.cuda import synchronize
from torch.utils.cpp_extension import load
from .config import PPQ_CONFIG
from .defs import ppq_warning, SingletonMeta


class ComplieHelper(metaclass=SingletonMeta):
    """ PPQ-Torch Compile Wrapper. """
    def __init__(self) -> None:
        self.__CUDA_EXTENTION__ = None

    def complie(self):
        ppq_warning('Compling Kernels... Please wait (It will take a few minutes).')
        lock_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'csrc/build/lock')
        if os.path.exists(lock_file): 
            try: os.remove(lock_file)
            except Exception as e:
                raise PermissionError(f'Can not delete lock file at {lock_file}, delete it first!')

        self.__CUDA_EXTENTION__ = load(
            name='PPQ_Cuda_Impls',
            sources=[
                os.path.join(os.path.dirname(os.path.dirname(__file__)), 'csrc/export.cc'),
                os.path.join(os.path.dirname(os.path.dirname(__file__)), 'csrc/cuda/linear.cu'),
                os.path.join(os.path.dirname(os.path.dirname(__file__)), 'csrc/cuda/sort.cu'),
                os.path.join(os.path.dirname(os.path.dirname(__file__)), 'csrc/cuda/train.cu'),
                os.path.join(os.path.dirname(os.path.dirname(__file__)), 'csrc/cuda/floating.cu'),
                os.path.join(os.path.dirname(os.path.dirname(__file__)), 'csrc/cpu/hist_mse.cc'),
            ],
            build_directory=os.path.join(os.path.dirname(os.path.dirname(__file__)), 'csrc/build/'),
            with_cuda=True,
            extra_cflags=['-O3'])
    
    @ property
    def CUDA_EXTENSION(self):
        if self.__CUDA_EXTENTION__ is None:
            raise Exception(
                'Cuda Extension has not been compiled, '
                'invoke ppq.core.ffi.ComplieHelper.complie() First.')
        return self.__CUDA_EXTENTION__

CUDA_COMPLIER = ComplieHelper()
if PPQ_CONFIG.USING_CUDA_KERNEL:
    CUDA_COMPLIER.complie()

# helper class for calling cuda methods.
class CUDA:
    """CUDA is a helper class for invoking highly-effcient custimized cuda
    kernel. PPQ developer team has implemented a series of quantization related
    cuda kernel, They are 5-100x faster than torch kernels, with less gpu
    memory cost.

    You can easily extend your cuda kernel via this class:
        Firstly, implement your kernel within ppq/csrc/cuda, write your own .cu file and .h file.
        Secondly, add your functions to ppq/csrc/cuda/export.cc, add them to export table.
        Finally, add a interface with this python class(ppq.core.ffi.CUDA),
        following the signature as same as others.

    PPQ CUDA EXTENSION 命名规则:
        我们使用函数名+后缀名的形式命名 CUDA Extension 函数:

        后缀名 _T 表示 Tensorwise 函数
        后缀名 _C 表示 Channelwise 函数
        后缀名 _B 表示 导函数

    例如函数 LinearQuantize_T_B 表示线性量化函数的 Tensorwise 版本，并且是导函数。
    """

    @ staticmethod
    def LinearQuantize_T(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        minimum: int = -128,
        maximum: int = 127,
        rounding: int = 0
    ) -> torch.Tensor:
        # if scale is too small, quantization might cause fp32 underflow.
        # if scale < 1e-7: raise ValueError('scale is too small.')
        return CUDA_COMPLIER.CUDA_EXTENSION.QuantizeTensor_LT(
            tensor, scales, offsets, minimum, maximum, rounding)

    @ staticmethod
    def LinearQuantize_C(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        channel_axis: int,
        minimum: int = -128,
        maximum: int = 127,
        rounding: int = 0
    ) -> torch.Tensor:
        return CUDA_COMPLIER.CUDA_EXTENSION.QuantizeTensor_LC(
            tensor, scales, offsets, minimum, maximum, channel_axis, rounding)

    @ staticmethod
    def LinearQuantize_T_B(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        dy: torch.Tensor,
        minimum: int,
        maximum: int,
        rounding: int,
    ) -> List[torch.Tensor]:
        return CUDA_COMPLIER.CUDA_EXTENSION.QuantizeTensor_LT_B(
            tensor, scales, offsets,
            dy, minimum, maximum, rounding
        )

    @ staticmethod
    def LinearQuantize_C_B(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        dy: torch.Tensor,
        minimum: int,
        maximum: int,
        channel_axis: int,
        rounding: int,
    ) -> List[torch.Tensor]:
        return CUDA_COMPLIER.CUDA_EXTENSION.QuantizeTensor_LC_B(
            tensor, scales, offsets,
            dy, minimum, maximum, rounding, channel_axis
        )

    @ staticmethod
    def Histogram_T(
        tensor: torch.Tensor,
        histogram: torch.Tensor,
        scale: float,
        clip_outliers: bool = True
    ) -> torch.Tensor:
        # if scale < 1e-7: raise ValueError('scale is too small.')
        CUDA_COMPLIER.CUDA_EXTENSION.Histogram_T(tensor, scale, clip_outliers, histogram)
        return histogram

    @ staticmethod
    def Histogram_Asymmetric_T(
        min_value: float,
        max_value: float,
        tensor: torch.Tensor,
        histogram: torch.Tensor,
        clip_outliers: bool = True
    ) -> torch.Tensor:
        # if scale < 1e-7: raise ValueError('scale is too small.')
        CUDA_COMPLIER.CUDA_EXTENSION.Histogram_Asymmetric_T(min_value, max_value, tensor, clip_outliers, histogram)
        return histogram

    @ staticmethod
    def Histogram_C(
        tensor: torch.Tensor,
        channel_axis: int,
        histogram: torch.Tensor,
        scale: float,
        clip_outliers: bool = True
    ) -> torch.Tensor:
        # if scale < 1e-7: raise ValueError('scale is too small.')
        CUDA_COMPLIER.CUDA_EXTENSION.Histogram_C(tensor, channel_axis, scale, clip_outliers, histogram)
        return histogram

    @ staticmethod
    def Quantile(
        tensor: torch.Tensor,
        q: float,
    ) -> torch.Tensor:
        return CUDA_COMPLIER.CUDA_EXTENSION.Quantile_T(tensor, q)

    @ staticmethod
    def TensorClip_T(
        tensor: torch.Tensor,
        reference: torch.Tensor,
        limit: torch.Tensor,
    ) -> torch.Tensor:
        if not tensor.is_contiguous(): tensor = tensor.contiguous()
        if not reference.is_contiguous(): tensor = reference.contiguous()
        return CUDA_COMPLIER.CUDA_EXTENSION.TensorClip_T(tensor, reference, limit)

    @ staticmethod
    def TensorClip_C(
        tensor: torch.Tensor,
        reference: torch.Tensor,
        limit: torch.Tensor,
        channel_axis: int,
    ) -> torch.Tensor:
        if not tensor.is_contiguous(): tensor = tensor.contiguous()
        if not reference.is_contiguous(): tensor = reference.contiguous()
        return CUDA_COMPLIER.CUDA_EXTENSION.TensorClip_C(
            tensor, reference, limit, channel_axis)

    @ staticmethod
    def RoundingLoss_LT(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        minimum: int = -128,
        maximum: int = 127,
        rounding: int = 0
    ) -> torch.Tensor:
        if not tensor.is_contiguous(): tensor = tensor.contiguous()
        return CUDA_COMPLIER.CUDA_EXTENSION.RoundingLoss_LT(
            tensor, scales, offsets, minimum, maximum, rounding)

    @ staticmethod
    def RoundingLoss_LT_B(
        tensor: torch.Tensor,
        dy    : torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        minimum: int = -128,
        maximum: int = 127,
        rounding: int = 0
    ) -> torch.Tensor:
        if not tensor.is_contiguous(): tensor = tensor.contiguous()
        return CUDA_COMPLIER.CUDA_EXTENSION.RoundingLoss_LT_B(
            tensor, dy, scales, offsets, minimum, maximum, rounding)

    @ staticmethod
    def RoundingLoss_LC(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        channel_axis: int,
        minimum: int = -128,
        maximum: int = 127,
        rounding: int = 0
    ) -> torch.Tensor:
        if not tensor.is_contiguous(): tensor = tensor.contiguous()
        return CUDA_COMPLIER.CUDA_EXTENSION.RoundingLoss_LC(
            tensor, scales, offsets, minimum, maximum, channel_axis, rounding)

    @ staticmethod
    def RoundingLoss_LC_B(
        tensor: torch.Tensor,
        dy    : torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        channel_axis: int,
        minimum: int = -128,
        maximum: int = 127,
        rounding: int = 0
    ) -> torch.Tensor:
        if not tensor.is_contiguous(): tensor = tensor.contiguous()
        return CUDA_COMPLIER.CUDA_EXTENSION.RoundingLoss_LC_B(
            tensor, dy, scales, offsets, minimum, maximum, channel_axis, rounding)

    @ staticmethod
    def OrderPreservingObserve(
        tensor: torch.Tensor,
    ) -> torch.Tensor:
        if not tensor.is_contiguous(): tensor = tensor.contiguous()
        return CUDA_COMPLIER.CUDA_EXTENSION.RoundingLoss_LC_B(tensor)

    @ staticmethod
    def compute_mse_loss(
        histogram: list,
        start: int,
        step: int,
        end: int
    ) -> float:
        return CUDA_COMPLIER.CUDA_EXTENSION.compute_mse_loss(histogram, start, step, end)

    @ staticmethod
    def FloatingQuantize_T(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        exponent: int = 4, 
        mantissa: int = 3,
        minimum: float = - 448, # FP8 E4M3
        maximum: float = + 448,
        rounding: int = 0
    ) -> torch.Tensor:
        if exponent <= 0: raise ValueError('Floating Quantization requires exponent > 0')
        if not tensor.is_contiguous(): tensor = tensor.contiguous()
        # if scale is too small, quantization might cause fp32 underflow.
        # if scale < 1e-7: raise ValueError('scale is too small.')
        return CUDA_COMPLIER.CUDA_EXTENSION.QuantizeTensor_FT(
            tensor, scales, offsets, exponent, mantissa, minimum, maximum, rounding)

    @ staticmethod
    def FloatingQuantize_C(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        channel_axis: int,
        exponent: int = 4, 
        mantissa: int = 3,
        minimum: float = - 448, # FP8 E4M3
        maximum: float = + 448,
        rounding: int = 0
    ) -> torch.Tensor:
        if exponent <= 0: raise ValueError('Floating Quantization requires exponent > 0')
        if not tensor.is_contiguous(): tensor = tensor.contiguous()
        return CUDA_COMPLIER.CUDA_EXTENSION.QuantizeTensor_FC(
            tensor, scales, offsets, exponent, mantissa, 
            minimum, maximum, channel_axis, rounding)

    @ staticmethod
    def FloatingQuantize_T_B(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        dy: torch.Tensor,
        exponent: int, 
        mantissa: int,
        minimum: float,
        maximum: float,
        rounding: int,
    ) -> List[torch.Tensor]:
        if not tensor.is_contiguous(): tensor = tensor.contiguous()
        return CUDA_COMPLIER.CUDA_EXTENSION.QuantizeTensor_FT_B(
            tensor, scales, offsets,
            dy, exponent, mantissa, minimum, maximum, rounding
        )

    @ staticmethod
    def FloatingQuantize_C_B(
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        dy: torch.Tensor,
        exponent: int, 
        mantissa: int,
        minimum: float,
        maximum: float,
        channel_axis: int,
        rounding: int,
    ) -> List[torch.Tensor]:
        if not tensor.is_contiguous(): tensor = tensor.contiguous()
        return CUDA_COMPLIER.CUDA_EXTENSION.QuantizeTensor_FC_B(
            tensor, scales, offsets,
            dy, exponent, mantissa, minimum, maximum, 
            rounding, channel_axis
        )


    @ staticmethod
    def Sync():
        """Synchronize device."""
        synchronize()



from .core import *
from .executor import BaseGraphExecutor, TorchExecutor, TorchQuantizeDelegator
from .IR import (
    BaseGraph,
    GraphBuilder,
    GraphCommand,
    GraphExporter,
    GraphFormatter,
    Operation,
    QuantableGraph,
    SearchableGraph,
    Variable,
    TrainableGraph,
)
from .scheduler import AggresiveDispatcher, ConservativeDispatcher, GraphDispatcher, PPLNNDispatcher
from .scheduler.perseus import Perseus



import time
from abc import ABCMeta, abstractmethod
from typing import Container, Iterable, Iterator, List

from ...executor import BaseGraphExecutor
from ...IR import BaseGraph, BaseGraph


class QuantizationOptimizationPass(metaclass=ABCMeta):
    """QuantizationOptimizationPass is a basic building block of PPQ
    quantization logic.

    PPQ is designed as a Multi pass Compiler of quantization network.
        where pass here refers to a traversal through the entire network.

    This class is an abstract base class of all customized passes.
    Quantizer will build an optimization pipeline later to quantize and optimize your network.
    """

    def __init__(self, name: str = "Default Quanzation Optim") -> None:
        self.name = name

    @abstractmethod
    def optimize(self, graph: BaseGraph, **kwargs) -> None:
        raise NotImplementedError("Implement this function first.")

    def __str__(self) -> str:
        return f"QuantizationOptimizationPass[{self.name}]"


class QuantizationOptimizationPipeline(Container, Iterable):
    """QuantizationOptimizationPipeline is a sorted set PPQ Optimization
    passes.

    PPQ is designed as a Multi pass Compiler of quantization network.
        where pass here refers to a traversal through the entire network.

    Quantizer is going to calling optimization pass from pipeline one by one to
        eventually finish network quantization procedure
    """

    def __init__(self, passes: List[QuantizationOptimizationPass]):
        super().__init__()
        self._pipeline = []
        if passes is not None:
            for optim in passes:
                self.append_optimization_to_pipeline(optim_pass=optim)

    def __len__(self) -> int:
        return len(self._pipeline)

    def __contains__(self, __x: QuantizationOptimizationPass) -> bool:
        assert isinstance(__x, QuantizationOptimizationPass), (
            f"Quantization Optimization Pipeline object only suppose to contain optimization passes, "
            f"while you require to check a/an {type(__x)} whether in the optimization list"
        )
        return __x in self._pipeline

    def __iter__(self) -> Iterator[QuantizationOptimizationPass]:
        return self._pipeline.__iter__()

    def optimize(self, graph: BaseGraph, verbose: bool = True, **kwargs) -> None:
        max_name_length = 0
        if len(self._pipeline) > 0:
            names = [p.name for p in self._pipeline]
            max_name_length = max([len(name) for name in names])

        for optim_pass in self._pipeline:
            if not isinstance(optim_pass, QuantizationOptimizationPass):
                raise TypeError(
                    f"Quantization Optimization Pipeline object only suppose to contain optimization passes only, "
                    f"while {str(optim_pass)}({type(optim_pass)}) was found."
                )

            if verbose:
                padding_length = abs(max_name_length - len(optim_pass.name))
                print(
                    f'[{time.strftime("%H:%M:%S", time.localtime())}] {optim_pass.name} Running ... '
                    + " " * padding_length,
                    end="",
                )

            if not isinstance(graph, BaseGraph):
                raise TypeError(
                    f"parameter 1 should be an instance of PPQ BaseGraph when calling optim pass, "
                    f"however {type(graph)} was given."
                )
            optim_pass.optimize(graph=graph, **kwargs)
            if verbose:
                print(f"Finished.")

    def append_optimization_to_pipeline(self, optim_pass: QuantizationOptimizationPass, at_front: bool = False):
        assert isinstance(optim_pass, QuantizationOptimizationPass), (
            f"Quantization Optimization Pipeline object only suppose to contain optimization passes, "
            f"while we got a/an {type(optim_pass)} in the optimization list"
        )
        if not at_front:
            self._pipeline.append(optim_pass)
        else:
            self._pipeline = [optim_pass] + self._pipeline
        return self

    def report(self) -> str:
        report = ""
        for optimization_pass in self._pipeline:
            report += str(optimization_pass) + "\n"
        return report



from enum import Enum
from math import sqrt
from typing import List

import torch
from ...IR import Operation


class EqualizationMethod(Enum):
    """

    EqualizationMethod - 用于列举所有可行的 scale 计算方案
                        different arithmetic methods for equalization
    """

    # key value = np.max(np.abs(x))
    ABSOLUTE_MAX = (1,)
    # key value = np.mean(np.abs(x))
    ABSOLUTE_MEAN = (2,)
    # key value = np.max(np.square(x))
    SQUARE_MAX = (3,)
    # key value = np.mean(np.square(x))
    SQUARE_MEAN = (4,)

    DIRECT_MAX = 5


class EqualizationHelper:
    @staticmethod
    def key_value_from_upstream(
        op: Operation,
        including_weight: bool = True,
        weight_multiplier: float = 1.0,
        including_bias: bool = False,
        including_act: bool = False,
        bias_multiplier: float = 0.5,
        act_multiplier: float = 0.5,
    ) -> torch.Tensor:
        if op.type not in {"Gemm", "MatMul", "Conv", "ConvTranspose"}:
            raise TypeError(f"Unsupported Op type {op.name}({op.type}) for Equalization Optimization.")
        if not op.inputs[1].is_parameter:
            raise ValueError(f"Parameter of Op {op.name} is non-static.")
        buffer = []

        # ----------------------------------
        # step - 1, extract weight from op:
        # ----------------------------------
        if including_weight:
            w = op.inputs[1].value * weight_multiplier
            if op.type == "ConvTranspose":
                num_of_groups = op.attributes.get("group", 1)
                if w.ndim == 3:
                    w = torch.reshape(w, (num_of_groups, w.shape[0] // num_of_groups) + w.shape[1:])
                    w = torch.permute(w, (2, 0, 1, 3))
                    w = torch.reshape(w, (w.shape[0] * w.shape[1], -1))
                elif w.ndim == 4:
                    w = torch.reshape(w, (num_of_groups, w.shape[0] // num_of_groups) + w.shape[1:])
                    w = torch.permute(w, (2, 0, 1, 3, 4))
                    w = torch.reshape(w, (w.shape[0] * w.shape[1], -1))
                elif w.ndim == 5:
                    w = torch.reshape(w, (num_of_groups, w.shape[0] // num_of_groups) + w.shape[1:])
                    w = torch.permute(w, (2, 0, 1, 3, 4, 5))
                    w = torch.reshape(w, (w.shape[0] * w.shape[1], -1))
                else:
                    raise ValueError(f"Unexpected dimension of weight of {op.name}.")
                buffer.append(w)

        if op.type in {"MatMul", "Gemm"}:
            assert w.ndim == 2, f"Unexpected Error, Parameter of MatMul {op.name} should be 2-d."
            if op.attributes.get("transB", 0) == 0:
                w = torch.transpose(w, 1, 0)
            buffer.append(w)

        if op.type == "Conv":
            w = torch.reshape(w, (w.shape[0], -1))
            buffer.append(w)

        # ----------------------------------
        # step - 2, extract bias from op:
        # ----------------------------------
        if including_bias and op.num_of_input == 3:
            b = op.inputs[-1].value * bias_multiplier
            if op.type in {"Conv", "Gemm"} and op.inputs[-1].is_parameter:
                b = torch.reshape(b, (w.shape[0], 1))
                buffer.append(b)

            if op.type == "ConvTranspose":
                b = torch.reshape(b, (w.shape[0], 1))
                buffer.append(b)

        # ----------------------------------
        # step - 3, extract activation from op:
        # ----------------------------------
        if including_act and op.outputs[0].value is not None:
            a = op.outputs[0].value * act_multiplier
            buffer.append(a)

        # concat and return
        return torch.cat(buffer, dim=-1)

    @staticmethod
    def key_value_from_downstream(op: Operation, weight_multiplier: float = 1.0) -> torch.Tensor:
        # ----------------------------------
        # step - 1, extract weight from op:
        # ----------------------------------
        w = op.inputs[1].value * weight_multiplier
        if op.type == "ConvTranspose":
            w = torch.reshape(w, (w.shape[0], -1))

        if op.type in {"MatMul", "Gemm"}:
            assert w.ndim == 2, f"Unexpected Error, Parameter of MatMul {op.name} should be 2-d."
            if op.attributes.get("transB", 0) != 0:
                w = torch.transpose(w, 1, 0)

        if op.type == "Conv":
            # for group convolution, we have to select its weight by group
            num_of_groups = op.attributes.get("group", 1)
            if w.ndim == 3:
                w = torch.reshape(w, (num_of_groups, w.shape[0] // num_of_groups) + w.shape[1:])
                if hasattr(torch, "permute"):
                    w = torch.permute(w, (2, 0, 1, 3))
                else:
                    w = w.permute((2, 0, 1, 3))
                w = torch.reshape(w, (w.shape[0] * w.shape[1], -1))
            elif w.ndim == 4:
                w = torch.reshape(w, (num_of_groups, w.shape[0] // num_of_groups) + w.shape[1:])
                if hasattr(torch, "permute"):
                    w = torch.permute(w, (2, 0, 1, 3, 4))
                else:
                    w = w.permute((2, 0, 1, 3, 4))
                w = torch.reshape(w, (w.shape[0] * w.shape[1], -1))
            elif w.ndim == 5:
                w = torch.reshape(w, (num_of_groups, w.shape[0] // num_of_groups) + w.shape[1:])
                if hasattr(torch, "permute"):
                    w = torch.permute(w, (2, 0, 1, 3, 4, 5))
                else:
                    w = w.permute((2, 0, 1, 3, 4, 5))
                w = torch.reshape(w, (w.shape[0] * w.shape[1], -1))
            else:
                raise ValueError(f"Unexpected dimension of weight of {op.name}.")
        return w

    @staticmethod
    def scale_to_upstream(op: Operation, scale_factor: torch.Tensor):
        if op.type not in {"Gemm", "MatMul", "Conv", "ConvTranspose"}:
            raise TypeError(f"Unsupported Op type {op.name}({op.type}) for Equalization Optimization.")
        if not op.inputs[1].is_parameter:
            raise ValueError(f"Parameter of Op {op.name} is non-static.")

        w = op.inputs[1].value
        has_bias = op.num_of_input == 3
        if has_bias and not op.inputs[-1].is_parameter:
            raise ValueError(f"Bias of Op {op.name} is non-static.")
        if has_bias:
            bias = op.inputs[-1].value

        if op.type == "ConvTranspose":
            num_of_groups = op.attributes.get("group", 1)
            w = torch.reshape(w, (num_of_groups, w.shape[0] // num_of_groups) + w.shape[1:])
            w *= torch.reshape(scale_factor, [num_of_groups, 1, -1] + [1] * (w.ndim - 3))
            w = torch.reshape(w, (w.shape[0] * w.shape[1],) + w.shape[2:])
            if has_bias:
                bias *= scale_factor

        elif op.type == "Conv":
            w *= torch.reshape(scale_factor, [-1] + ([1] * (w.ndim - 1)))
            if has_bias:
                bias *= scale_factor

        elif op.type in {"Gemm", "MatMul"}:
            if op.attributes.get("transB", 0) == 0:
                w = torch.transpose(w, 1, 0)
            w *= torch.reshape(scale_factor, (-1, 1))
            if op.attributes.get("transB", 0) == 0:
                w = torch.transpose(w, 1, 0)
            if has_bias:
                bias *= scale_factor

        # write back
        with torch.no_grad():
            op.inputs[1].value.copy_(w)
            if has_bias:
                op.inputs[-1].value.copy_(bias)

    @staticmethod
    def scale_to_downstream(op: Operation, scale_factor: torch.Tensor):
        if op.type not in {"Gemm", "MatMul", "Conv", "ConvTranspose"}:
            raise TypeError(f"Unsupported Op type {op.name}({op.type}) for Equalization Optimization.")
        if not op.inputs[1].is_parameter:
            raise ValueError(f"Parameter of Op {op.name} is non-static.")
        w = op.inputs[1].value

        if op.type == "ConvTranspose":
            w /= torch.reshape(scale_factor, [-1] + ([1] * (w.ndim - 1)))

        if op.type == "Conv":
            num_of_groups = op.attributes.get("group", 1)
            w = torch.reshape(w, (num_of_groups, w.shape[0] // num_of_groups) + w.shape[1:])
            w /= torch.reshape(scale_factor, [num_of_groups, 1, -1] + [1] * (w.ndim - 3))
            w = torch.reshape(w, (w.shape[1] * num_of_groups,) + w.shape[2:])

        if op.type in {"Gemm", "MatMul"}:
            if op.attributes.get("transB", 0) != 0:
                w = torch.transpose(w, 1, 0)
            w /= torch.reshape(scale_factor, (-1, 1))
            if op.attributes.get("transB", 0) != 0:
                w = torch.transpose(w, 1, 0)

        # write back
        with torch.no_grad():
            op.inputs[1].value.copy_(w)


class ChannelSplitHelper:
    @staticmethod
    def split_by_mask(mask: torch.Tensor, tensor: torch.Tensor, axis: int, scale_factor: float):
        mask = mask.tolist()
        if tensor.shape[axis] != len(mask):
            raise ValueError(f"Unexpected tensor shape {tensor.shape}. Expect {len(mask)} at axis {axis}.")
        tensor = tensor.transpose(dim0=axis, dim1=0)

        concat_view = []
        for idx, _mask in enumerate(mask):
            if _mask == 1:
                concat_view.append(tensor[idx].unsqueeze(0) * scale_factor)
                concat_view.append(tensor[idx].unsqueeze(0) * scale_factor)
            else:
                concat_view.append(tensor[idx].unsqueeze(0))

        return torch.cat(concat_view, dim=0).transpose(dim0=0, dim1=axis)

    @staticmethod
    def channel_split_upstream(op: Operation, mask: torch.Tensor, scale_factor: float):
        if op.type not in {"Gemm", "MatMul", "Conv", "ConvTranspose"}:
            raise TypeError(f"Unsupported Op type {op.name}({op.type}) for Channel Split Optimization.")
        if not op.inputs[1].is_parameter:
            raise ValueError(f"Parameter of Op {op.name} is non-static.")

        w = op.inputs[1].value
        has_bias = op.num_of_input == 3
        if has_bias and not op.inputs[-1].is_parameter:
            raise ValueError(f"Bias of Op {op.name} is non-static.")
        if has_bias:
            bias = op.inputs[-1].value

        if op.type == "ConvTranspose":
            num_of_groups = op.attributes.get("group", 1)
            w = torch.reshape(w, (num_of_groups, w.shape[0] // num_of_groups) + w.shape[1:])
            w = torch.transpose(w, 1, 2)  # swap output channel and input channel
            w = torch.reshape(w, (-1,) + w.shape[2:])  # merge group and input channel
            w = ChannelSplitHelper.split_by_mask(mask=mask, tensor=w, axis=0, scale_factor=scale_factor)
            w = torch.reshape(w, (num_of_groups, -1) + w.shape[1:])
            w = torch.transpose(w, 1, 2)  # swap output channel and input channel
            w = torch.reshape(w, (w.shape[1] * num_of_groups,) + w.shape[2:])
            if has_bias:
                bias = ChannelSplitHelper.split_by_mask(mask=mask, tensor=bias, axis=0, scale_factor=scale_factor)

        elif op.type == "Conv":
            w = ChannelSplitHelper.split_by_mask(mask=mask, tensor=w, axis=0, scale_factor=scale_factor)
            if has_bias:
                bias = ChannelSplitHelper.split_by_mask(mask=mask, tensor=bias, axis=0, scale_factor=scale_factor)
            num_of_groups = op.attributes.get("group", 1)

        elif op.type in {"Gemm", "MatMul"}:
            if op.attributes.get("transB", 0) == 0:
                w = torch.transpose(w, 1, 0)
            w = ChannelSplitHelper.split_by_mask(mask=mask, tensor=w, axis=0, scale_factor=scale_factor)
            if op.attributes.get("transB", 0) == 0:
                w = torch.transpose(w, 1, 0)
            if has_bias:
                bias = ChannelSplitHelper.split_by_mask(mask=mask, tensor=bias, axis=0, scale_factor=scale_factor)

        # write back
        with torch.no_grad():
            op.inputs[1].value = w
            if has_bias:
                op.inputs[-1].value = bias

    def channel_split_downstream(op: Operation, mask: torch.Tensor, scale_factor: float):
        if op.type not in {"Gemm", "MatMul", "Conv", "ConvTranspose"}:
            raise TypeError(f"Unsupported Op type {op.name}({op.type}) for Channel Split Optimization.")
        if not op.inputs[1].is_parameter:
            raise ValueError(f"Parameter of Op {op.name} is non-static.")
        w = op.inputs[1].value

        if op.type == "ConvTranspose":
            w = ChannelSplitHelper.split_by_mask(mask=mask, tensor=w, axis=0, scale_factor=scale_factor)

        if op.type == "Conv":
            num_of_groups = op.attributes.get("group", 1)
            w = torch.reshape(w, (num_of_groups, w.shape[0] // num_of_groups) + w.shape[1:])
            w = torch.transpose(w, 1, 2)  # swap output channel and input channel
            w = torch.reshape(w, (-1,) + w.shape[2:])  # merge group and input channel
            w = ChannelSplitHelper.split_by_mask(mask=mask, tensor=w, axis=0, scale_factor=scale_factor)
            w = torch.reshape(w, (num_of_groups, -1) + w.shape[1:])
            w = torch.transpose(w, 1, 2)  # swap output channel and input channel
            w = torch.reshape(w, (w.shape[1] * num_of_groups,) + w.shape[2:])

        if op.type in {"Gemm", "MatMul"}:
            if op.attributes.get("transB", 0) != 0:
                w = torch.transpose(w, 1, 0)
            w = ChannelSplitHelper.split_by_mask(mask=mask, tensor=w, axis=0, scale_factor=scale_factor)
            if op.attributes.get("transB", 0) != 0:
                w = torch.transpose(w, 1, 0)

        # write back
        with torch.no_grad():
            op.inputs[1].value = w


class EqualizationPair:
    def __init__(self, upstream_layers: List[Operation], downstream_layers: List[Operation]):
        """
            EqualizationPair - 一个数据结构，封装了 equalization 的核心数据抽象和执行逻辑
                               a class encapsulating execution logic of equalization

                在 self.upstream_layers 包含了 equalization 操作中的所有上游层(op)
                self.upstream_layers contain all upstream ops

                在 self.downstream_layers 包含了 equalization 操作中的所有下游层(op)
                self.downstream_layers contain all downstream ops

            一个 EqualizationPair 结构记录了参与 equalization 变换的所有相关层与其局部图结构信息
            从而支持在局部子图上的 equalization 操作
            An EqualizationPair records all relevant ops participating in the equalization
            transformation, thus supporting equalization on local subgraphs
        Args:
            upstream_layers (list):
                equalization 操作中的所有上游层(op)

            downstream_layers (list):
                equalization 操作中的所有下游层(op)
        """
        self.upstream_layers = upstream_layers
        self.downstream_layers = downstream_layers

    def equalize(
        self,
        value_threshold: float,
        including_weight: bool = True,
        weight_multiplier: float = 1.0,
        including_act: bool = False,
        act_multiplier: float = 0.5,
        including_bias: bool = False,
        bias_multiplier: float = 0.5,
        method: EqualizationMethod = EqualizationMethod.ABSOLUTE_MAX,
    ):
        # extract key value from pair
        upstream_key_values, downstream_key_values = [], []
        for op in self.upstream_layers:
            key_value = EqualizationHelper.key_value_from_upstream(
                op=op,
                including_bias=including_bias,
                including_act=including_act,
                bias_multiplier=bias_multiplier,
                act_multiplier=act_multiplier,
            )
            upstream_key_values.append(key_value)

        for op in self.downstream_layers:
            key_value = EqualizationHelper.key_value_from_downstream(op=op)
            downstream_key_values.append(key_value)

        upstream_key_values = self.reduce_by_axis(upstream_key_values, method=method)
        downstream_key_values = self.reduce_by_axis(downstream_key_values, method=method)

        # calculate scale
        scale = self.calculate_scale(
            upstream_key_values=upstream_key_values,
            downstream_key_values=downstream_key_values,
            value_threshold=value_threshold,
        )

        # write back all params
        for op in self.upstream_layers:
            EqualizationHelper.scale_to_upstream(op, scale)

        for op in self.downstream_layers:
            EqualizationHelper.scale_to_downstream(op, scale)

    def channel_split(
        self,
        value_threshold: float = 2,
        including_act: bool = False,
        act_multiplier: float = 0.5,
        including_bias: bool = False,
        bias_multiplier: float = 0.5,
        method: EqualizationMethod = EqualizationMethod.ABSOLUTE_MAX,
    ):
        # extract key value from pair
        upstream_key_values, downstream_key_values = [], []
        for op in self.upstream_layers:
            key_value = EqualizationHelper.key_value_from_upstream(
                op=op,
                including_bias=including_bias,
                including_act=including_act,
                bias_multiplier=bias_multiplier,
                act_multiplier=act_multiplier,
            )
            upstream_key_values.append(key_value)

        for op in self.downstream_layers:
            key_value = EqualizationHelper.key_value_from_downstream(op=op)
            downstream_key_values.append(key_value)

        upstream_key_values = self.reduce_by_axis(upstream_key_values, method=method)
        downstream_key_values = self.reduce_by_axis(downstream_key_values, method=method)

        mask = torch.logical_and(upstream_key_values >= value_threshold, downstream_key_values >= value_threshold)

        # write back all params
        for op in self.upstream_layers:
            ChannelSplitHelper.channel_split_upstream(op=op, scale_factor=1 / sqrt(2), mask=mask)

        for op in self.downstream_layers:
            ChannelSplitHelper.channel_split_downstream(op=op, scale_factor=1 / sqrt(2), mask=mask)

    def activation_equalize(self, threshold: float = 4.0):
        # extract key value from pair
        upstream_key_values = []
        for op in self.upstream_layers:
            key_value = EqualizationHelper.key_value_from_upstream(
                op=op,
                including_bias=False,
                including_act=True,
                bias_multiplier=0,
                act_multiplier=1,
                weight_multiplier=0,
            )
            upstream_key_values.append(key_value)
        upstream_key_values = self.reduce_by_axis(upstream_key_values, method=EqualizationMethod.ABSOLUTE_MAX)

        threshold = upstream_key_values.mean().item()
        # calculate scale
        scale = torch.where(
            upstream_key_values > threshold,
            torch.ones_like(upstream_key_values) * 1.1,
            torch.ones_like(upstream_key_values) / 1.1,
        )

        # write back all params
        for op in self.upstream_layers:
            EqualizationHelper.scale_to_upstream(op, 1 / scale)

        for op in self.downstream_layers:
            EqualizationHelper.scale_to_downstream(op, 1 / scale)

    def calculate_scale(
        self,
        upstream_key_values: torch.Tensor,
        downstream_key_values: torch.Tensor,
        value_threshold: float,
        scale_clip_value: float = 10,
    ):
        scale = 1 / torch.sqrt(upstream_key_values / downstream_key_values)
        scale = torch.clamp(scale, 1 / scale_clip_value, scale_clip_value)
        scale[(upstream_key_values + downstream_key_values) < value_threshold] = 1
        return scale

    def reduce_by_axis(
        self,
        params: List[torch.Tensor],
        method: EqualizationMethod,
        axis: int = 1,
    ) -> torch.Tensor:
        params = torch.cat(params, axis=axis)
        if method is EqualizationMethod.ABSOLUTE_MAX:
            return torch.max(torch.abs(params), axis=axis)[0]

        elif method is EqualizationMethod.ABSOLUTE_MEAN:
            return torch.mean(torch.abs(params), axis=axis)

        elif method is EqualizationMethod.SQUARE_MAX:
            return torch.max(torch.square(params), axis=axis)[0]

        elif method is EqualizationMethod.SQUARE_MEAN:
            return torch.mean(torch.square(params), axis=axis)

        elif method is EqualizationMethod.DIRECT_MAX:
            return torch.max(params, axis=axis)

        else:
            raise NotImplementedError("Equalization method %s is not support." % str(method))



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
import copy
import os
import pathlib
from collections import OrderedDict, deque
from tempfile import TemporaryDirectory
from typing import Dict, List, Optional, Sequence, Union

import onnx
import onnx_graphsurgeon as osg
from onnxruntime.tools.onnx_model_utils import (get_optimization_level,
                                                optimize_model)

from xslim.defs import MIN_ONNX_OPSET_VERSION
from xslim.logger import logger

from .onnxslim_pass import infer_onnx_model, optimize_onnx_model


def get_onnx_opset(onnx_model: onnx.ModelProto) -> Dict[str, int]:
    opset_dict = {}
    for opset in onnx_model.opset_import:
        _domain = opset.domain
        _domain = "ai.onnx" if _domain == "" else _domain
        opset_dict[_domain] = opset.version

    return opset_dict


def ensure_default_onnx_opset(onnx_model: onnx.ModelProto, min_onnx_version: int = MIN_ONNX_OPSET_VERSION) -> int:
    ai_onnx_version = None
    for opset in onnx_model.opset_import:
        if opset.domain in {"", "ai.onnx"}:
            ai_onnx_version = opset.version
            break

    if ai_onnx_version is None:
        logger.warning(
            f"Missing default ONNX opset import, defaulting to {min_onnx_version}.")
        opset = onnx_model.opset_import.add()
        opset.domain = ""
        opset.version = min_onnx_version
        ai_onnx_version = min_onnx_version
    return ai_onnx_version


def _normalize_kernel_shape_attrs(
    onnx_model: onnx.ModelProto,
) -> onnx.ModelProto:
    try:
        osg_graph = osg.import_onnx(onnx_model)
        updated = False

        for node in osg_graph.nodes:
            if node.op not in {"Conv", "ConvTranspose"}:
                continue
            if "kernel_shape" in node.attrs:
                continue
            if len(node.inputs) < 2:
                continue

            weight_var = node.inputs[1]
            weight_shape = None
            if isinstance(weight_var, osg.Constant):
                weight_values = getattr(weight_var, "values", None)
                if weight_values is not None and hasattr(weight_values, "shape"):
                    weight_shape = list(weight_values.shape)

            if weight_shape is None:
                inferred_shape = getattr(weight_var, "shape", None)
                if inferred_shape is not None:
                    weight_shape = list(inferred_shape)

            if weight_shape is None or len(weight_shape) < 3:
                logger.warning(
                    (
                        "skip filling kernel_shape for %s because "
                        "weight shape is unavailable"
                    ),
                    node.name or node.op,
                )
                continue

            kernel_shape = weight_shape[2:]
            if any(dim is None or isinstance(dim, str) for dim in kernel_shape):
                logger.warning(
                    (
                        "skip filling kernel_shape for %s because "
                        "weight shape %s is not static"
                    ),
                    node.name or node.op,
                    weight_shape,
                )
                continue

            node.attrs["kernel_shape"] = [int(dim) for dim in kernel_shape]
            updated = True

        if not updated:
            return onnx_model

        return osg.export_onnx(osg_graph)
    except Exception as exc:
        logger.warning(
            "Failed to normalize kernel_shape attributes via GraphSurgeon; "
            "returning original ONNX model. Error: %s",
            exc,
        )
        return onnx_model
def format_onnx_model(
    onnx_model: onnx.ModelProto, sim_en: bool = True, min_onnx_version: int = MIN_ONNX_OPSET_VERSION
) -> onnx.ModelProto:
    """
    Regularize an onnx model, including removing shape fields, value_info fields, etc., to avoid entering bugs.

    Args:
        onnx_model (onnx.ModelProto): input onnx Model
        min_onnx_version (int, optional): min onnx opset version. Defaults to 13.

    Returns:
        onnx.ModelProto: output ONNX Model
    """
    onnx_model.graph.ClearField("value_info")
    for o_var in onnx_model.graph.output:
        try:
            for dim in o_var.type.tensor_type.shape.dim:
                dim.dim_value = 0
                dim.dim_param = "?"
        except:
            pass

    empty_name_count = 0
    for node in onnx_model.graph.node:
        if node.name == "":
            node.name = f"{node.op_type}_{empty_name_count}"
            empty_name_count += 1

    onnx_model = _normalize_kernel_shape_attrs(onnx_model)

    ai_onnx_version = ensure_default_onnx_opset(onnx_model, min_onnx_version)
    if ai_onnx_version < min_onnx_version:
        logger.warning("convert ai.onnx version {} to {}...".format(
            ai_onnx_version, min_onnx_version))
        onnx_model = onnx.version_converter.convert_version(
            onnx_model, min_onnx_version)

    if sim_en:
        logger.info("simplify onnx model...")
    try:
        onnx_model = optimize_onnx_model(onnx_model)
    except Exception as e:
        logger.warning("simplify onnx model error and skip. {}".format(e))

    try:
        onnx_model = infer_onnx_model(onnx_model)
    except Exception as e:
        logger.warning("shape_inference error with {}, skiped".format(e))

    return onnx_model


def merge_onnx_model(
    onnx_model: onnx.ModelProto,
    truncate_left_graph: Optional[osg.Graph] = None,
    truncate_vars: Optional[Sequence[osg.Variable]] = None,
):
    if isinstance(truncate_left_graph, osg.Graph) and isinstance(truncate_vars, Sequence):
        osg_graph = osg.import_onnx(onnx_model)
        for idx, o_var in enumerate(osg_graph.outputs):
            o_idx = o_var.inputs[0].outputs.index(o_var)
            o_var.inputs[0].outputs[o_idx] = truncate_vars[idx]

        new_osg_graph = osg.Graph(
            nodes=osg_graph.nodes + truncate_left_graph.nodes,
            inputs=truncate_left_graph.inputs,
            outputs=truncate_left_graph.outputs,
            name=copy.copy(osg_graph.name),
            doc_string=copy.copy(osg_graph.doc_string),
            opset=copy.copy(osg_graph.opset),
            import_domains=osg_graph.import_domains,
        )
        onnx_model = osg.export_onnx(new_osg_graph)

    return onnx_model


def truncate_onnx_model(onnx_model: onnx.ModelProto, truncate_var_names: Optional[Sequence[str]] = None):
    if isinstance(truncate_var_names, Sequence) and len(truncate_var_names) > 0:
        if len(set(truncate_var_names)) != len(truncate_var_names):
            raise RuntimeError(
                "The incoming truncate_var_names contains duplicate tensor names")
        truncate_vars = []
        osg_graph = osg.import_onnx(onnx_model)
        tensors = osg_graph.tensors()
        for k, v in tensors.items():
            if k in set(truncate_var_names):
                truncate_vars.append(v)

        graph_valid_truncate_var_names = set([t.name for t in truncate_vars])

        invalid_var_names = set(truncate_var_names) ^ set(
            graph_valid_truncate_var_names)
        if len(invalid_var_names) > 0:
            raise RuntimeError(
                "The incoming truncate_var_names contains non-existent tensor names {}".format(
                    ", ".join(invalid_var_names)
                )
            )

        valid_node_names = set()
        invalid_node_names = set()
        graph_node_names = set()
        dst_node_dict = {}
        src_node_dict = {}
        for i, node in enumerate(osg_graph.nodes):
            if node.name == "":
                node.name = "{}_{}_{}".format(node.op, i, id(node))
            if node.name in graph_node_names:
                node.name = "{}_{}_{}".format(node.name, i, id(node))
            graph_node_names.add(node.name)
            dst_node_dict[node.name] = set()
            src_node_dict[node.name] = set()
            for var in node.outputs:
                for dst_node in var.outputs:
                    dst_node_dict[node.name].add(dst_node.name)
            for var in node.inputs:
                for src_node in var.inputs:
                    src_node_dict[node.name].add(src_node.name)

        def _truncate_graph_upstream(out_vars: Sequence[osg.Tensor]):
            visit_ops = deque()

            def _upstream_impl(vars: Sequence[osg.Tensor]):
                for var in vars:
                    for source_op in var.inputs:
                        if source_op.name in valid_node_names or source_op.name not in graph_node_names:
                            continue
                        valid_node_names.add(source_op.name)
                        visit_ops.append(source_op)

            _upstream_impl(out_vars)
            while len(visit_ops) > 0:
                dq_size = len(visit_ops)
                for _ in range(dq_size):
                    up_op = visit_ops.popleft()
                    _upstream_impl(up_op.inputs)

        def _truncate_graph_downstream(out_vars: Sequence[osg.Tensor]):
            visit_ops = deque()

            def _upstream_impl(vars: Sequence[osg.Tensor]):
                for var in vars:
                    for source_op in var.inputs:
                        if (
                            source_op.name in invalid_node_names
                            or source_op.name in valid_node_names
                            or source_op.name not in graph_node_names
                        ):
                            continue
                        invalid_node_names.add(source_op.name)
                        visit_ops.append(source_op)

            def _downstream_impl(vars: Sequence[osg.Tensor]):
                for var in vars:
                    for dest_op in var.outputs:
                        if dest_op.name in invalid_node_names or dest_op.name not in graph_node_names:
                            continue
                        invalid_node_names.add(dest_op.name)
                        visit_ops.append(dest_op)

            _downstream_impl(out_vars)
            while len(visit_ops) > 0:
                dq_size = len(visit_ops)
                for _ in range(dq_size):
                    up_op = visit_ops.popleft()
                    _downstream_impl(up_op.outputs)
                    _upstream_impl(up_op.inputs)

        _truncate_graph_upstream(truncate_vars)
        _truncate_graph_downstream(truncate_vars)

        if len(invalid_node_names) + len(valid_node_names) != len(graph_node_names):
            raise RuntimeError("truncate graph failed.")

        valid_nodes = []
        invalid_nodes = []

        for node in osg_graph.nodes:
            if node.name in valid_node_names:
                valid_nodes.append(node)
            elif node.name in invalid_node_names:
                invalid_nodes.append(node)
            else:
                raise RuntimeError(
                    "unexpected error for node {}".format(node.name))

        truncate_graph = osg.Graph(
            nodes=valid_nodes,
            inputs=osg_graph.inputs,
            outputs=truncate_vars,
            name=copy.copy(osg_graph.name),
            doc_string=copy.copy(osg_graph.doc_string),
            opset=copy.copy(osg_graph.opset),
            import_domains=osg_graph.import_domains,
        )

        truncate_left_graph = osg.Graph(
            nodes=invalid_nodes,
            inputs=osg_graph.inputs,
            outputs=osg_graph.outputs,
            name=copy.copy(osg_graph.name),
            doc_string=copy.copy(osg_graph.doc_string),
            opset=copy.copy(osg_graph.opset),
            import_domains=osg_graph.import_domains,
        )

        truncate_onnx_model = osg.export_onnx(truncate_graph)

        for var in truncate_vars:
            var.inputs.clear()

        return truncate_onnx_model, truncate_left_graph, truncate_vars
    else:
        return onnx_model, None, None



import numpy as np
import onnxslim.third_party.onnx_graphsurgeon as osg
from onnxslim.core.pattern import Pattern, PatternMatcher
from onnxslim.core.pattern.registry import register_fusion_pattern
from xslim.logger import logger


class BatchNormDecompositionMatcher(PatternMatcher):
    def __init__(self, priority):
        pattern = Pattern(
            """
            input      X        0 1 bn
            BatchNormalization bn 5 1 X ? ? ? ? output
            output     output   1 0 bn
            """
        )
        super().__init__(pattern, priority)

    @property
    def name(self):
        return "DecomposeBatchNormalization"

    def rewrite(self, opset=11):
        bn_node = self.bn
        X = bn_node.inputs[0]
        scale_const = bn_node.inputs[1]
        B_const = bn_node.inputs[2]
        mean_const = bn_node.inputs[3]
        var_const = bn_node.inputs[4]
        output_var = bn_node.outputs[0]

        if not all(isinstance(t, osg.Constant) for t in [scale_const, B_const, mean_const, var_const]):
            return {}

        scale = scale_const.values
        B = B_const.values
        mean = mean_const.values
        var = var_const.values
        epsilon = bn_node.attrs.get("epsilon", 0.000009999999747378752)

        try:
            denom = np.sqrt(var + epsilon)
            a_1d = scale / denom
            b_1d = B - a_1d * mean
            x_shape = X.shape
            if not x_shape:
                logger.warning("[Warning] No shape info, skip BN optimize.")
                return {}
            else:
                axis = 1
                broadcast_shape = [1] * len(x_shape)
                broadcast_shape[axis] = -1
                a = a_1d.reshape(broadcast_shape)
                b = b_1d.reshape(broadcast_shape)
        except Exception as e:
            logger.warning(f"[Warning] Failed to compute a/b for BN {bn_node.name}: {e}")
            return {}

        a_const = osg.Constant(name=f"{bn_node.name}_a", values=a)
        b_const = osg.Constant(name=f"{bn_node.name}_b", values=b)

        mul_output = osg.Variable(name=f"{bn_node.name}_mul_out", dtype=X.dtype)

        new_ops = {}
        X.outputs.remove(bn_node)
        output_var.inputs.clear()

        new_ops[f"{bn_node.name}_mul"] = {
            "op": "Mul",
            "inputs": [X, a_const],
            "outputs": [mul_output],
            "domain": None,
            "attrs": {},
            "name": f"{bn_node.name}_mul",
        }

        new_ops[f"{bn_node.name}_add"] = {
            "op": "Add",
            "inputs": [mul_output, b_const],
            "outputs": [output_var],
            "domain": None,
            "attrs": {},
            "name": f"{bn_node.name}_add",
        }

        return new_ops

register_fusion_pattern(BatchNormDecompositionMatcher(1))


from typing import Tuple

import torch
from xslim.defs import OBSERVER_MIN_SCALE_THRESHOLD
from xslim.logger import logger

from ...core import (
    CUDA,
    OBSERVER_KL_COMPUTING_DEVICE,
    OBSERVER_KL_HIST_BINS,
    OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE,
    OBSERVER_MIN_SCALE,
    OBSERVER_MIN_SCALE_MANUL_OVERRIDE,
    OBSERVER_MSE_COMPUTE_INTERVAL,
    OBSERVER_MSE_HIST_BINS,
    OBSERVER_PERCENTILE,
    OBSERVER_PERCENTILE_MANUL_OVERRIDE,
    OBSERVER_WARNING,
    PPQ_CONFIG,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    TensorQuantizationConfig,
    convert_any_to_numpy,
    ppq_quant_param_computing_function,
)
from ...IR import Variable
from ...utils.round import ppq_numerical_round, ppq_round_to_power_of_2
from ..measure import torch_KL_divergence
from .base import BaseTensorObserver


@ppq_quant_param_computing_function
def minmax_to_scale_offset(
    min_val: float,
    max_val: float,
    config: TensorQuantizationConfig,
    scale_threshold: float = OBSERVER_MIN_SCALE_THRESHOLD,
) -> Tuple[float, float]:
    """
    Solve scale and offset with given min, max value.
    For Symmetrical Quantization, offset is set to 0.
    For ASymmetrical Quantization, offset is limited by [config.quant_min, config.quant_max].

    Scale is limited by [scale_threshold, +inf].

    Args:
        min_val (float): min value
        max_val (float): max value
        config (TensorQuantizationConfig): Corresponding TQC.
        scale_threshold (float, optional): minimum scale.

    Returns:
        Tuple[float, float]: Solved scale and offset.
    """
    if OBSERVER_MIN_SCALE_MANUL_OVERRIDE in config.detail:
        scale_threshold = config.detail[OBSERVER_MIN_SCALE_MANUL_OVERRIDE]

    scale, offset = 1, 0
    if min_val > 0:
        min_val = 0
    if max_val < 0:
        max_val = 0

    if config.policy.has_property(QuantizationProperty.ASYMMETRICAL):
        range = float(max_val - min_val)
        scale = range / (config.quant_max - config.quant_min)
        scale = max(scale, scale_threshold)
        offset = config.quant_min - ppq_numerical_round(min_val / scale)

    elif config.policy.has_property(QuantizationProperty.SYMMETRICAL):
        range = 2 * float(max(abs(max_val), abs(min_val)))
        scale = range / (config.quant_max - config.quant_min)
        scale = max(scale, scale_threshold)
        offset = 0

    else:
        raise TypeError("Tensor Min Max Observer Excepts either ASYMMETRICAL or SYMMETRICAL quantization config.")
    if config.policy.has_property(QuantizationProperty.POWER_OF_2):
        scale = ppq_round_to_power_of_2(scale, policy=RoundingPolicy.ROUND_UP)
    return scale, offset


class TorchMinMaxObserver(BaseTensorObserver):
    """TorchMinMaxObserver collects min and max value of given tensor."""

    def __init__(self, watch_on: Variable, quant_cfg: TensorQuantizationConfig):
        super().__init__(watch_on, quant_cfg)
        self._min_val_collector = []
        self._max_val_collector = []
        self._value_device = "cpu"
        self._none_value_detected = False

    @torch.no_grad()
    def observe(self, value: torch.Tensor):
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return
        assert isinstance(value, torch.Tensor), "TorchMinMaxObserver can only deal with torch Tensor values"
        self._value_device = value.device.type
        if value is None or value.numel() == 0:
            self._none_value_detected = True
            return
        if self._quant_cfg.state == QuantizationStates.INITIAL:
            if self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
                self._min_val_collector.append(
                    value.min().reshape(
                        shape=[
                            1,
                        ]
                    )
                )
                self._max_val_collector.append(
                    value.max().reshape(
                        shape=[
                            1,
                        ]
                    )
                )
            elif self._quant_cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
                channel_axis = self._quant_cfg.channel_axis
                channelwise_view = value.transpose(dim0=0, dim1=channel_axis).unsqueeze(-1)
                channelwise_view = torch.flatten(channelwise_view, start_dim=1)
                self._min_val_collector.append(torch.min(channelwise_view, dim=1, keepdim=True)[0])
                self._max_val_collector.append(torch.max(channelwise_view, dim=1, keepdim=True)[0])
            else:
                raise TypeError("Min-max Observer only work with per-tensor or per-channel quantize policy.")

    def render_quantization_config(self):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        if self._none_value_detected:
            logger.warning("None value detected at var {}".format(self._watch_on.name))
            self._quant_cfg.scale = torch.tensor([1.0], dtype=torch.float32, device=self._value_device).squeeze(0)
            self._quant_cfg.offset = torch.tensor([0], dtype=torch.float32, device=self._value_device).squeeze(0)
            self._quant_cfg.state = QuantizationStates.ACTIVATED
            self._quant_cfg.detail["NONE_VALUE"] = True
            return

        if len(self._max_val_collector) == 0:
            raise ValueError(
                "Can not render quantization config yet, Observer data collator is empty. "
                "Invoke observe() function before render config."
            )
        device = self._max_val_collector[-1].device

        if self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
            scale, offset = minmax_to_scale_offset(
                min_val=torch.min(torch.cat(self._min_val_collector, dim=0)).cpu().item(),
                max_val=torch.max(torch.cat(self._max_val_collector, dim=0)).cpu().item(),
                config=self._quant_cfg,
            )
            self._quant_cfg.scale = torch.tensor([scale], dtype=torch.float32, device=device).squeeze(0)
            self._quant_cfg.offset = torch.tensor([offset], dtype=torch.float32, device=device).squeeze(0)
            self._quant_cfg.state = QuantizationStates.ACTIVATED

        elif self._quant_cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
            min_vals = torch.min(torch.cat(self._min_val_collector, dim=-1), dim=-1, keepdim=False)[0].cpu().numpy()
            max_vals = torch.max(torch.cat(self._max_val_collector, dim=-1), dim=-1, keepdim=False)[0].cpu().numpy()
            assert len(min_vals) == len(max_vals), "Min values and max values should at same length."
            scales, offsets = [], []
            for min_val, max_val in zip(min_vals, max_vals):
                scale, offset = minmax_to_scale_offset(min_val=min_val, max_val=max_val, config=self._quant_cfg)
                scales.append(scale)
                offsets.append(offset)

            # scale, offset here only deployed on cpu
            # we will move them towards target device through RunnableGraph
            self._quant_cfg.scale = torch.tensor(scales, dtype=torch.float32, device=device)
            self._quant_cfg.offset = torch.tensor(offsets, dtype=torch.float32, device=device)
            self._quant_cfg.state = QuantizationStates.ACTIVATED
        else:
            raise TypeError("Min-max Observer only work with per-tensor or per-channel quantize policy.")


class TorchHistObserver(TorchMinMaxObserver):
    """TorchHistObserver collects histogram of given tensor.

    It is designed for per-tensor quantization or activation quantization.
    """

    def __init__(self, watch_on: Variable, quant_cfg: TensorQuantizationConfig, hist_bins: int = OBSERVER_KL_HIST_BINS):
        self._phase = "Detecting Minmax"
        self._hist = None
        self._hist_scale = None
        self._min = None
        self._max = None
        if OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE in quant_cfg.detail:
            hist_bins = quant_cfg.detail[OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE]
        self._hist_bins = hist_bins
        super().__init__(watch_on, quant_cfg)

    def observe(self, value: torch.Tensor):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        assert value.numel() > 0, f"You are observing an empty tensor({self._watch_on.name})."

        if self._phase == "Detecting Minmax":
            return super().observe(value)  # collect min, max

        elif self._phase == "Collating Hist":
            if self._hist is None:
                self._hist = torch.zeros(size=(self._hist_bins,), dtype=torch.int32, device=value.device)

            if self._quant_cfg.policy.has_property(QuantizationProperty.ASYMMETRICAL):
                # ASYMMETRICAL Hist
                if PPQ_CONFIG.USING_CUDA_KERNEL and value.is_cuda:
                    CUDA.Histogram_Asymmetric_T(self._min, self._max, tensor=value, histogram=self._hist)
                else:
                    hist = torch.histc(value, self._hist_bins, min=self._min, max=self._max)
                    self._hist += hist.int()

            elif self._quant_cfg.policy.has_property(QuantizationProperty.SYMMETRICAL):
                # SYMMETRICAL Hist
                if PPQ_CONFIG.USING_CUDA_KERNEL and value.is_cuda:
                    CUDA.Histogram_T(tensor=value, histogram=self._hist, scale=self._hist_scale)
                else:
                    hist = torch.histc(torch.abs(value), self._hist_bins, min=0, max=self._hist_scale * self._hist_bins)
                    self._hist += hist.int()

            else:
                raise TypeError(
                    "Quantization Property is invalid, " "expect either ASYMMETRICAL or SYMMETRICAL config here."
                )

    @ppq_quant_param_computing_function
    def hist_to_scale_offset(
        self,
        histogram: torch.Tensor,
        hist_bins: int,
        hist_scale: float,
        config: TensorQuantizationConfig,
        computing_device: str = OBSERVER_KL_COMPUTING_DEVICE,
        scale_threshold: float = OBSERVER_MIN_SCALE,
    ) -> Tuple[float, int]:
        """
        PPQ core quant parameter computing method - Histogram to scale & offset

        With a pre-defined histogram,
        this function will automatically search best clip value
        to minimize KL divergence between quantized result and fp32 input.

        only work for per-tensor symmetrical quantization policy for now.
        see also https://on-demand.gputechconf.com/gtc/2017/presentation/s7310-8-bit-inference-with-tensorrt.pdf

        Args:
            histogram (torch.Tensor): histogram records activation's statistics.
            hist_bins (int): how many bins are included in histogram(also known as histogram length)
            hist_scale (float): histogram step size. it can be solved by histogram.max_val / histogram.bins
            config (TensorQuantizationConfig): quantization config.
            computing_device (str, optional): computing device. Defaults to 'cpu'.

        Raises:
            ValueError: given quantization config is invalid.

        Returns:
            Tuple[float, int]: scale(fp32) and offset(int).
        """
        if config.policy.has_property(QuantizationProperty.ASYMMETRICAL):
            raise PermissionError("KL observer is not designed for ASYMMETRICAL quantization")

        if OBSERVER_MIN_SCALE_MANUL_OVERRIDE in config.detail:
            scale_threshold = config.detail[OBSERVER_MIN_SCALE_MANUL_OVERRIDE]

        # move histogram to cpu, speedup computation.
        histogram = histogram.to(computing_device).float()

        # compute symmtrical kl-divergence.
        # Here is a simple example: reference distribution P consisting of 8 bins, we want to quantize into 2 bins:
        # P = [ 1, 0, 2, 3, 5, 3, 1, 7]
        # we merge into 2 bins (8 / 2 = 4 consecutive bins are merged into one bin)
        # [1 + 0 + 2 + 3 , 5 + 3 + 1 + 7] = [6, 16]
        # then proportionally expand back to 8 bins, we preserve empty bins from the original distribution P:
        # Q = [ 6/3, 0, 6/3, 6/3, 16/4, 16/4, 16/4, 16/4] = [ 2, 0, 2, 2, 4, 4, 4, 4]
        # now we should normalize both distributions, after that we can compute KL_divergence
        # P /= sum(P) Q /= sum(Q)
        # result = KL_divergence(P, Q)
        # see also
        # https://github.com/NVIDIA/TensorRT/blob/3835424af081db4dc8cfa3ff3c9f4a8b89844421/tools/pytorch-quantization/pytorch_quantization/calib/histogram.py#L147

        losses, quant_bins = [], 2 ** (config.num_of_bits - 1)

        # following code is curcial, do not remove
        histogram[: int(hist_bins * 0.002)] = 0
        histogram[int(hist_bins * 0.002)] = 1

        hist_sum = torch.sum(histogram)
        for bin_range in range(quant_bins, hist_bins + quant_bins - 1, quant_bins):
            p_hist = torch.zeros(size=(bin_range,), dtype=torch.float, device=computing_device)
            p_hist[:bin_range].copy_(histogram[:bin_range])
            p_hist[bin_range - 1] += torch.sum(histogram[bin_range:])
            p_hist = p_hist / hist_sum

            expand_ratio = int(bin_range / quant_bins)
            q_hist = histogram[:bin_range].clone()
            q_hist = q_hist.reshape((quant_bins, expand_ratio))
            positive_map = q_hist > 0
            positive_cnt = positive_map.sum(axis=1, keepdim=True)
            positive_cnt[positive_cnt == 0] = 1
            q_hist = torch.div(q_hist.sum(axis=1, keepdim=True), positive_cnt)
            q_hist = q_hist.repeat([1, expand_ratio])
            q_hist = q_hist * positive_map
            q_hist = q_hist / torch.sum(q_hist)
            q_hist = q_hist.flatten()

            losses.append({"kl": torch_KL_divergence(p_hist, q_hist), "bin_range": bin_range})

        best_bin_range = sorted(losses, key=lambda x: x["kl"])[0]["bin_range"]
        scale, offset = (best_bin_range / self._hist_bins) * hist_scale * (self._hist_bins / quant_bins), 0

        if scale < scale_threshold and OBSERVER_WARNING:
            logger.warning(
                "Numeric instability detected: "
                "ppq find there is a scale value < 1e-7, "
                "which probably cause numeric underflow in further computation."
            )
        scale = max(scale, scale_threshold)

        if config.policy.has_property(QuantizationProperty.POWER_OF_2):
            scale = ppq_round_to_power_of_2(scale, policy=RoundingPolicy.ROUND_HALF_UP)
        return scale, offset

    def render_quantization_config(self):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        if not self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
            raise ValueError("Hist observer can only apply with per-tensor quantization config.")

        if self._phase == "Detecting Minmax":
            min_val = torch.min(torch.cat(self._min_val_collector, dim=0)).cpu().item()
            max_val = torch.max(torch.cat(self._max_val_collector, dim=0)).cpu().item()
            if self._quant_cfg.policy.has_property(QuantizationProperty.SYMMETRICAL):
                hist_range = float(max(abs(max_val), abs(min_val)))
            else:
                hist_range = max_val - min_val
            self._min = min_val
            self._max = max_val
            self._hist_scale = hist_range / self._hist_bins
            self._phase = "Collating Hist"
        elif self._phase == "Collating Hist":
            scale, offset = self.hist_to_scale_offset(
                histogram=self._hist, hist_bins=self._hist_bins, hist_scale=self._hist_scale, config=self._quant_cfg
            )
            device = self._hist.device
            self._quant_cfg.scale = torch.tensor([scale], dtype=torch.float32, device=device).squeeze(0)
            self._quant_cfg.offset = torch.tensor([offset], dtype=torch.float32, device=device).squeeze(0)
            self._quant_cfg.state = QuantizationStates.ACTIVATED


class TorchPercentileObserver(BaseTensorObserver):
    """TorchPercentileObserver collects percentile data of given tensor.

    It is designed for activation quantization.
    """

    def __init__(self, watch_on: Variable, quant_cfg: TensorQuantizationConfig):
        super().__init__(watch_on, quant_cfg)
        if not OBSERVER_PERCENTILE_MANUL_OVERRIDE in quant_cfg.detail:
            self._percentile = OBSERVER_PERCENTILE
        else:
            self._percentile = quant_cfg.detail[OBSERVER_PERCENTILE_MANUL_OVERRIDE]
        self._percentile_collector = []
        self._percentile_maxs = []
        self._percentile_mins = []
        self._value_device = "cpu"
        self._none_value_detected = False

    @torch.no_grad()
    def observe(self, value: torch.Tensor):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        assert isinstance(value, torch.Tensor), "TorchMinMaxObserver can only deal with torch Tensor values"
        if isinstance(value, torch.Tensor):
            self._value_device = value.device.type

        if value is None or value.numel() == 0:
            self._none_value_detected = True
            return
        if self._quant_cfg.state == QuantizationStates.INITIAL:
            if self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
                if not PPQ_CONFIG.USING_CUDA_KERNEL or (not value.is_cuda):
                    numel = value.numel()

                    min_idx, max_idx = int(numel * (1 - self._percentile)), int(numel * (self._percentile))
                    # torch.kthvalue needs index from 1 to numel ...
                    min_idx = max(0, min_idx) + 1
                    max_idx = min(max_idx, numel - 1) + 1
                    _min = torch.kthvalue(value.flatten(), k=min_idx, dim=0)[0].view(1, -1)
                    _max = torch.kthvalue(value.flatten(), k=max_idx, dim=0)[0].view(1, -1)
                    self._percentile_collector.append(torch.cat([_max, _min], dim=-1))
                else:
                    self._percentile_collector.append(CUDA.Quantile(value, self._percentile).view(1, -1))
            else:
                raise TypeError("Min-max Observer only work with per-tensor quantize policy.")

    def render_quantization_config(self):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        if self._none_value_detected:
            logger.warning("None value detected at var {}".format(self._watch_on.name))
            self._quant_cfg.scale = torch.tensor([1.0], dtype=torch.float32, device=self._value_device).squeeze(0)
            self._quant_cfg.offset = torch.tensor([0], dtype=torch.float32, device=self._value_device).squeeze(0)
            self._quant_cfg.state = QuantizationStates.ACTIVATED
            self._quant_cfg.detail["NONE_VALUE"] = True
            return

        if self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
            if len(self._percentile_collector) == 0:
                raise ValueError(
                    "Can not render quantization config yet, Observer data collator is empty. "
                    "Invoke observe() function before render config."
                )
            device = self._percentile_collector[-1].device
            self._percentile_collector = torch.cat(self._percentile_collector, dim=0).float().mean(dim=0).cpu()
            scale, offset = minmax_to_scale_offset(
                min_val=self._percentile_collector[1].item(),
                max_val=self._percentile_collector[0].item(),
                config=self._quant_cfg,
            )

            self._quant_cfg.scale = torch.tensor([scale], dtype=torch.float32, device=device).squeeze(0)
            self._quant_cfg.offset = torch.tensor([offset], dtype=torch.float32, device=device).squeeze(0)
            self._quant_cfg.state = QuantizationStates.ACTIVATED
        elif self._quant_cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
            raise PermissionError("Percentile observer can not deal with per channel quantization.")
            if len(self._percentile_maxs) == 0:
                raise ValueError(
                    "Can not render quantization config yet, Observer data collator is empty. "
                    "Invoke observe() function before render config."
                )
            device = self._percentile_maxs[-1].device

            min_vals = torch.mean(torch.cat(self._percentile_mins, dim=-1), dim=-1, keepdim=False)
            max_vals = torch.mean(torch.cat(self._percentile_maxs, dim=-1), dim=-1, keepdim=False)

            min_vals = min_vals.cpu()
            max_vals = max_vals.cpu()

            assert len(min_vals) == len(max_vals), "Min values and max values should at same length."
            scales, offsets = [], []
            for min_val, max_val in zip(min_vals, max_vals):
                scale, offset = minmax_to_scale_offset(min_val=min_val, max_val=max_val, config=self._quant_cfg)
                scales.append(scale)
                offsets.append(offset)

            self._quant_cfg.scale = torch.tensor(scales, dtype=torch.float32, device=device)
            self._quant_cfg.offset = torch.tensor(offsets, dtype=torch.int32, device=device)
            self._quant_cfg.state = QuantizationStates.ACTIVATED
        else:
            raise TypeError("Min-max Observer only work with per-tensor or per-channel quantize policy.")


class TorchMSEObserver(TorchHistObserver):
    """Histogram accelerated MSE Observer, inspired by LightGBM This observer
    will collect data in histogram firstly, all mse computing will directly use
    histogram rather than data itself.

        Time complexity: O(Iteration * Num_of_Batch * Length(Data)) -> O(Iteration * Length(histogram))
        Space complexity: O(Num_of_Batch * Length(Data)) -> O(Iteration * Length(histogram))

    Args:
        TorchHistObserver ([type]): [description]
    """

    def __init__(self, watch_on: Variable, quant_cfg: TensorQuantizationConfig, bins: int = OBSERVER_MSE_HIST_BINS):
        super().__init__(watch_on, quant_cfg)
        self._hist_bins = bins

    def compute_mse_loss(self, histogram: list, start: int, step: int, end: int):
        if PPQ_CONFIG.USING_CUDA_KERNEL:
            raise NotImplementedError()
        else:
            # 如果你觉得 mse 太慢，想办法加速这段代码就可以了
            # 求解 mse 时，我们假设每一个 bin 里面的数据都是均匀分布的
            # 我们需要给一个直方图，并用 start, end, step 给出量化表示的范围
            # losses = [0 for _ in histogram]  debug
            num_of_elements = sum(histogram)
            loss = 0
            for idx, bin in enumerate(histogram):
                if idx < start:
                    # 如果所选的 bin 已经超出了起点，那从 bin 的中心到起点的距离即
                    # ((idx 到 起点的距离) + 0.5)
                    # 注意 hist 统计时是直接取 floor 的，因此会在这里额外 - 1
                    error = (start - idx - 1) + 0.5
                elif idx > end:
                    # 注意 hist 统计时是直接取 floor 的
                    error = (idx - end) + 0.5
                else:
                    # 分别计算左右两侧的 err
                    l_idx = (idx - start) % step
                    r_idx = step - l_idx - 1
                    if l_idx == r_idx:
                        error = l_idx + 0.25
                    else:
                        l_err = l_idx + 0.5
                        r_err = r_idx + 0.5
                        error = min(l_err, r_err)
                loss += (bin * error * error) / num_of_elements
                # losses[idx] = bin * error * error
            return loss

    @ppq_quant_param_computing_function
    def hist_to_scale_offset(
        self,
        histogram: torch.Tensor,
        hist_bins: int,
        hist_scale: float,
        config: TensorQuantizationConfig,
        scale_threshold: float = OBSERVER_MIN_SCALE,
    ) -> Tuple[float, int]:
        if OBSERVER_MIN_SCALE_MANUL_OVERRIDE in config.detail:
            scale_threshold = config.detail[OBSERVER_MIN_SCALE_MANUL_OVERRIDE]
        histogram = convert_any_to_numpy(histogram).tolist()
        num_of_quant_levels = (self._quant_cfg.quant_max - self._quant_cfg.quant_min) + 1

        losses = []
        if config.policy.has_property(QuantizationProperty.ASYMMETRICAL) and config.policy.has_property(
            QuantizationProperty.PER_TENSOR
        ):
            # at least we can have a min-max result
            step = hist_bins // num_of_quant_levels + 1
            loss = self.compute_mse_loss(histogram=histogram, start=0, step=step, end=num_of_quant_levels * step)
            losses.append({"mse": loss, "start": 0, "end": num_of_quant_levels * step})

            for start in range(0, hist_bins, OBSERVER_MSE_COMPUTE_INTERVAL):
                if (start * hist_scale) + self._min > 0:
                    break  # start can not > 0

                for step in range(1, hist_bins // num_of_quant_levels + 1):
                    end = start + num_of_quant_levels * step
                    if end > (hist_bins + num_of_quant_levels):
                        break
                    loss = self.compute_mse_loss(histogram=histogram, start=start, step=step, end=end)
                    losses.append({"mse": loss, "start": start, "end": end})

            best_policy = sorted(losses, key=lambda x: x["mse"])[0]
            best_start = best_policy["start"]
            best_end = best_policy["end"]

            # translate start & end to scale & offset.
            range_min, range_max = (best_start * hist_scale) + self._min, (best_end * hist_scale) + self._min
            scale, offset = minmax_to_scale_offset(range_min, range_max, config, scale_threshold)
            return scale, offset

        elif config.policy.has_property(QuantizationProperty.PER_CHANNEL):
            raise PermissionError("Torch Mse observer do not support PER_CHANNEL policy now, please wait.")

        elif config.policy.has_property(QuantizationProperty.SYMMETRICAL) and config.policy.has_property(
            QuantizationProperty.PER_TENSOR
        ):
            # at least we can have a min-max result
            step = hist_bins // num_of_quant_levels + 1
            loss = self.compute_mse_loss(histogram=histogram, start=0, step=step, end=num_of_quant_levels * step)
            losses.append({"mse": loss, "end": num_of_quant_levels * step})

            for step in range(1, hist_bins // num_of_quant_levels + 1):
                end = num_of_quant_levels * step
                if end > (hist_bins + num_of_quant_levels):
                    break
                loss = self.compute_mse_loss(histogram=histogram, start=0, step=step, end=end)
                losses.append({"mse": loss, "end": end})

            best_policy = sorted(losses, key=lambda x: x["mse"])[0]
            best_end = best_policy["end"]

            # translate start & end to scale & offset.
            range_min, range_max = -(best_end * hist_scale), (best_end * hist_scale)
            scale, offset = minmax_to_scale_offset(range_min, range_max, config, scale_threshold)
            return scale, offset

        raise Exception("Oops, there might be some mistakes.")



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
from collections import OrderedDict
from typing import Callable, Dict, Iterator, List, Optional

import torch
from pandas import DataFrame
from tqdm import tqdm
from xslim.logger import logger

from ..defs import PASSIVE_OPERATIONS
from ..ppq_decorator.ppq.executor import RuntimeHook, TorchExecutor
from ..ppq_decorator.ppq.IR import BaseGraph, Operation, QuantableOperation, Variable
from ..ppq_decorator.ppq.quantization.measure import torch_cosine_similarity, torch_mean_square_error, torch_snr_error
from ..ppq_decorator.ppq.utils.fetch import batch_random_fetch, generate_torch_indexer, tensor_random_fetch


class DetailedRecorder(RuntimeHook):
    def __init__(self, operation: Operation, fetchs: int = 1024) -> None:
        self.fetchs = fetchs
        self.i_storage = [[] for _ in range(operation.num_of_input)]
        self.o_storage = [[] for _ in range(operation.num_of_output)]
        self.i_indexer = [None for _ in range(operation.num_of_input)]
        self.o_indexer = [None for _ in range(operation.num_of_output)]
        super().__init__(operation)

    def pre_forward_hook(self, inputs: List[torch.Tensor], **kwargs) -> list:
        for idx, input in enumerate(inputs):
            if isinstance(input, torch.Tensor) and input.numel() > 0:
                if self.i_indexer[idx] is None:
                    self.i_indexer[idx] = generate_torch_indexer(self.fetchs, input.numel())
                self.i_storage[idx].append(input.flatten()[self.i_indexer[idx]])
            else:
                self.i_storage[idx].append(torch.ones([1]))
        return super().pre_forward_hook(inputs, **kwargs)

    def post_forward_hook(self, outputs: List[torch.Tensor], **kwargs) -> list:
        for idx, output in enumerate(outputs):
            if isinstance(output, torch.Tensor) and output.numel() > 0:
                if self.o_indexer[idx] is None:
                    self.o_indexer[idx] = generate_torch_indexer(self.fetchs, output.numel())
                self.o_storage[idx].append(output.flatten()[self.o_indexer[idx]])
            else:
                self.o_storage[idx].append(torch.ones([1]))
        return super().post_forward_hook(outputs, **kwargs)

    def clear(self):
        self.i_storage = [[] for _ in range(self._hook_to.num_of_input)]
        self.o_storage = [[] for _ in range(self._hook_to.num_of_output)]


def statistical_analyse(
    graph: BaseGraph,
    running_device: str,
    dataloader: Iterator,
    collate_fn: Callable = None,
    steps: int = 8,
    report_path: Optional[str] = None,
) -> List[dict]:
    class StatisticalErrorAnalyser:
        def __init__(self, x_fp: List[torch.Tensor], x_qt: List[torch.Tensor], op: Operation, var: Variable) -> None:
            self.x_qt = torch.cat(x_qt, dim=0)
            self.x_fp = torch.cat(x_fp, dim=0)
            self.x_er = self.x_qt - self.x_fp
            self.op = op
            self.var = var

            self.num_of_samples = self.x_fp.shape[0]

        def stat(self) -> dict:
            x_er, x_fp, x_qt = self.x_er, self.x_fp, self.x_qt
            er_mean = x_er.mean().item()
            er_std = x_er.std().item()

            qt_mean = x_qt.mean().item()
            qt_std = x_qt.std().item()
            qt_min = x_qt.min().item()
            qt_max = x_qt.max().item()

            fp_mean = x_fp.mean().item()
            fp_std = x_fp.std().item()
            fp_min = x_fp.min().item()
            fp_max = x_fp.max().item()
            fp_hist = torch.histc(x_fp, bins=32, min=x_fp.min(), max=x_fp.max()).cpu().tolist()

            snr = torch_snr_error(x_qt, x_fp).item()
            cosine = torch_cosine_similarity(x_qt, x_fp).item()
            mse = torch_mean_square_error(x_qt, x_fp).item()
            return {
                "SNR": snr,
                "MSE": mse,
                "Cosine": cosine,
                "Q.MinMax": "{:.3f}, {:.3f}".format(qt_min, qt_max),
                "F.MinMax": "{:.3f}, {:.3f}".format(fp_min, fp_max),
                "F.Hist": ",".join([str(int(i)) for i in fp_hist]),
                "is_parameter": self.var.is_parameter,
            }

    executor = TorchExecutor(graph=graph, device=running_device)
    # find all quantable operations.
    interested_op = []
    operation_list = graph.topological_sort()
    for operation in operation_list:
        if isinstance(operation, QuantableOperation) and operation.type not in PASSIVE_OPERATIONS:
            interested_op.append(operation)
    if len(interested_op) == 0:
        logger.warning("No analyzable operators were found.")

    # set up all hooks.
    hooks, caches = OrderedDict(), OrderedDict()
    for operation in interested_op:
        if isinstance(operation, QuantableOperation):
            hooks[operation.name] = DetailedRecorder(operation=operation, fetchs=2048)
            caches[operation.name] = {
                "Quantized Input": [],
                "Quantized Output": [],
                "Dequantized Input": [],
                "Dequantized Output": [],
            }

    # dequantize all
    for operation in operation_list:
        if isinstance(operation, QuantableOperation):
            operation.dequantize()

    # run for each quantable operations:
    analyse_data_list = []
    for idx, batch in enumerate(dataloader):
        if collate_fn is not None:
            batch = collate_fn(batch)
        analyse_data_list.append(batch)
        if idx >= steps:
            break

    for batch in tqdm(analyse_data_list, desc="Analysing Dequantized"):
        executor.forward(inputs=batch, hooks=hooks)

    for operation in interested_op:
        hook = hooks[operation.name]
        assert isinstance(hook, DetailedRecorder)
        caches[operation.name]["Dequantized Input"] = hook.i_storage.copy()
        caches[operation.name]["Dequantized Output"] = hook.o_storage.copy()
        hook.clear()

    # restore all
    for operation in operation_list:
        if isinstance(operation, QuantableOperation):
            operation.restore_quantize_state()

    # run for each quantable operations:
    for batch in tqdm(analyse_data_list, desc="Analysing Quantized"):
        executor.forward(inputs=batch, hooks=hooks)

    for operation in interested_op:
        hook = hooks[operation.name]
        assert isinstance(hook, DetailedRecorder)
        caches[operation.name]["Quantized Input"] = hook.i_storage.copy()
        caches[operation.name]["Quantized Output"] = hook.o_storage.copy()
        hook.clear()

    # analysing cache
    records = []
    visited_var = set()
    for name, record in caches.items():
        op_records = {}
        operation = graph.operations[name]
        if not isinstance(operation, Operation):
            continue
        op_records["Op"] = "{}[{}]".format(operation.name, operation.type)
        op_records["Vars"] = {}
        for idx, input_var in enumerate(operation.inputs):
            if input_var in visited_var:
                continue
            visited_var.add(input_var)
            x_qt = record["Quantized Input"][idx]
            x_fp = record["Dequantized Input"][idx]
            if x_fp[0].dtype not in {torch.float32, torch.float64, torch.float16}:
                continue
            op_records["Vars"][input_var.name] = StatisticalErrorAnalyser(
                x_fp=x_fp, x_qt=x_qt, op=operation, var=input_var
            ).stat()

        max_snr = 0
        for idx, output_var in enumerate(operation.outputs):
            if output_var in visited_var:
                continue
            visited_var.add(output_var)
            x_qt = record["Quantized Output"][idx]
            x_fp = record["Dequantized Output"][idx]
            if x_fp[0].dtype not in {torch.float32, torch.float64, torch.float16}:
                continue
            _detail = StatisticalErrorAnalyser(x_fp=x_fp, x_qt=x_qt, op=operation, var=output_var).stat()
            if _detail["SNR"] > max_snr:
                max_snr = _detail["SNR"]
            op_records["Vars"][output_var.name] = _detail

        op_records["MAX_SNR"] = max_snr
        records.append(op_records)

    def md_red_float(value):
        return '<font color="red">{:.4f}</font>'.format(value)

    sort_variable_reports = sorted(records, key=lambda x: x["MAX_SNR"], reverse=True)

    report_info_list = []
    for report_info in sort_variable_reports:
        for var_name, var_info in report_info["Vars"].items():
            report_info_list.append({"Op": report_info["Op"], "Var": var_name, **var_info})
            if "is_parameter" in report_info_list[-1]:
                is_parameter = report_info_list[-1].pop("is_parameter")
                if is_parameter:
                    report_info_list[-1]["Var"] = "{}[Constant]".format(report_info_list[-1]["Var"])
            snr_value = report_info_list[-1]["SNR"]
            cos_value = report_info_list[-1]["Cosine"]
            if snr_value > 0.1:
                report_info_list[-1]["SNR"] = md_red_float(snr_value)
            else:
                report_info_list[-1]["SNR"] = "{:.4f}".format(snr_value)

            if cos_value < 0.99:
                report_info_list[-1]["Cosine"] = md_red_float(cos_value)
            else:
                report_info_list[-1]["Cosine"] = "{:.4f}".format(cos_value)

            report_info_list[-1]["MSE"] = "{:.4f}".format(report_info_list[-1]["MSE"])

    report_index = ["Op", "Var", "SNR", "MSE", "Cosine", "Q.MinMax", "F.MinMax", "F.Hist"]
    report_info_df = DataFrame(report_info_list, columns=report_index)

    if isinstance(report_path, str):
        logger.info("export quantization statistical results file to {}".format(report_path))
        report_info_df.to_markdown(report_path)
    return report_info_list



from typing import Dict, List, Set

from ..core import TargetPlatform
from ..IR import BaseGraph, Operation, OpSocket, SearchableGraph

from .base import GraphDispatcher


class Perseus(GraphDispatcher):
    """英仙座 是 PPQ 0.6.5 版本引入的全新调度器

    基于新的 OpSocket, VProperty, VLink 数据类型抽象，
    英仙座调度器能够以一种可扩展的、可控制的方式切分你的计算图。

    英仙座调度器基于静态图分析与数值追踪,
    借由 OpSocket 定义的 "接线器" 算子模型, 英仙座调度器可以对
    Onnx中数据流的来源进行追踪，并进一步求解关于节点的数值传递闭包。
    这一过程完全是静态完成的，不需要送入数据执行。

    对于一个简单的 Reshape 算子而言，接线器模型定义了该算子内部的数据流动情况：
    (对于Reshape算子而言, 其输出结果与第一个输入之间存在数值关联, 而与第二个输入直接不存在直接关联)

      Input(VALUE)   Shape(SOI)
        |              |
    * ---------------------- *
    |   |                    |
    |   * ----- *   Reshape  |
    |           |            |
    * ---------------------- *
                |
               Out(VALUE)

    接线器模型在算子内部定义了这样的输入 - 输出关系网，从而调度器可以使用这样的关系网来划分子图。

    """

    def __init__(self, graph: BaseGraph, verbose: bool = False) -> None:
        """初始化英仙座调度器
        在这一过程中，英仙座调度器将为每一个算子生成它的 OpSocket 接线器
        一旦算子接线器初始化完毕，你将不能对图结构进行进一步修改，否则调度将会失败

        在这一过程中，英仙座将为 onnx 算子生成预定义的 onnx 算子接线器，
        所有被列举在 ppq/scheduler/core/default.py 中的算子均有预定义的算子接线器实现
        支持 opset 1 ~ 18

        如果你的网络具有自定义算子，或存在未在 default.py 列举的 onnx 算子
        英仙座会为他们生成默认接线器，请知悉：这可能导致错误

        你应当为你的新算子注册新的接线器，从而英仙座调度器能够使用新定义的接线器模型完成调度
        当英仙座调度器为你未知的算子生成默认接线器时，它将给出警报
        """
        self.sockets = {}
        self.verbose = verbose
        self.graph = graph
        self._search_engine = SearchableGraph(graph)
        self._precomputed_op_fanout = {}
        self._precomputed_op_fanin = {}

        for op in self.graph.operations.values():
            socket = op.socket
            fanin, fanout = set(), set()
            for link in socket.links:
                source_var = op.inputs[link.in_idx]
                dest_var = op.outputs[link.out_idx]
                source_op = source_var.source_op
                dest_ops = dest_var.dest_ops
                if source_op is not None:
                    fanin.add(source_op)
                for dest_op in dest_ops:
                    fanout.add(dest_op)
            self._precomputed_op_fanin[op.name], self._precomputed_op_fanout[op.name] = fanin, fanout
            self.sockets[op.name] = socket

        super().__init__()

    def solve_transitive_closure(self, sources: List[Operation], recursive: bool = True) -> Set[Operation]:
        """从指定节点出发解传递闭包，该方法是量化子图切分的核心函数
        传递闭包是一个节点的集合，以 C(x) = x_1, x_2, x_3 ... 进行表示
        以 C(x) 表示从节点 x 出发得到的传递闭包，则有以下结论
            1. C(x) 必须包含 x
            2. C(x) 中任意节点的传递闭包等于 C(x)

        在 Onnx 定义的神经网络中存在两类数据：
            其一是从 Gemm, Conv 等计算节点出发的数据，他们往往是需要量化的；
            以及从 Shape, TopK 出发的数据，他们往往是不能被量化的；

        为此 PPQ 将从网络中所有计算节点出发，求解关于计算节点的传递闭包 A
        而后从 Shape, TopK 等节点出发，求解非计算节点的传递闭包 B

        集合 A - B 中的节点将被量化，也被称为量化区节点
        集合 A * B 中的节点将被称为冲突区节点，默认不量化
        集合 B 中的节点将被称为 SOI 节点，不量化且调度到 Cpu 执行
        集合 A, B 之外的节点为未知区域节点，默认不量化

        PPQ 使用朴素递归求解传递闭包，对于神经网络这种稀疏图而言，

        其时间复杂度大约为O(n)

        注意一些细节问题，TopK 节点有两个输出：前K大的值和前K大的index，
        如果从 TopK 节点 出发求传递闭包，只会寻找到其第一个输出分支上的节点。

        节点上的链接关系由算子接线器进行定义；算子接线器是一种 PPQ 内置的抽象数据结构，
        参考：ppq.IR.base.opdef.py
        """
        if isinstance(sources, Operation):
            sources = [sources]

        closure = set(sources)
        closure_size = len(sources)
        closure_is_changing = True

        # loop until convergence
        while closure_is_changing:
            closure_is_changing = False
            b_extension = self.parse_transitive_fanin(closure)
            f_extension = self.parse_transitive_fanout(closure)
            closure.update(b_extension)
            closure.update(f_extension)
            if len(closure) != closure_size and recursive:
                closure_is_changing = True
            closure_size = len(closure)

        return set(closure)

    def mark_quantable_op(self) -> Set[Operation]:
        """追踪图中所有可量化节点，即求解所有计算节点的传递闭包"""
        sources = [op for op in self.graph.operations.values() if op.is_computing_op]
        return self.solve_transitive_closure(sources)

    def mark_soi_op(self) -> Set[Operation]:
        """追踪图中所有不可量化节点，即求解所有不可量化节点的传递闭包
        不可量化节点种类繁多，常见的节点包括 Shape, Topk, NMS
        有一些节点的输入不能量化，如 Reshape, Slice 等
        所有不可量化节点由 opsocket 具体定义给出
        """
        # initialize non quantable group.
        sources = set()
        for op in self.graph.operations.values():
            if op.name not in self.sockets:
                raise KeyError(f"Can not find Opsocket for {op.name}, graph has been modified.")
            socket = self.sockets[op.name]
            assert isinstance(socket, OpSocket)

            for plat, ovar in zip(socket.out_plat, op.outputs):
                if plat != TargetPlatform.UNSPECIFIED:
                    for dop, dix in zip(ovar.dest_ops, ovar.dest_idx):
                        if self.opsocket(dop).in_plat[dix] == TargetPlatform.UNSPECIFIED:
                            sources.add(dop)

            for plat, ivar in zip(socket.in_plat, op.inputs):
                if plat != TargetPlatform.UNSPECIFIED and ivar.source_op != None:
                    sources.add(ivar.source_op)
        return self.solve_transitive_closure(sources)

    def parse_transitive_fanout(self, parsing_from: List[Operation]) -> Set[Operation]:
        """从指定节点出发，寻找所有扇出节点
        注意此函数与 graph.get_downstream_operations 的区别
        graph.get_downstream_operations 函数将直接返回目标节点的所有下游节点
        而此函数将只返回下游节点中与当前节点存在数值传递关系的节点

        例如对于子图 conv - shape
            graph.get_downstream_operations(conv) 将返回 shape 节点
            parse_transitive_fanout(conv) 将返回空
        """
        fanout = self._search_engine.opset_matching(
            sp_expr=lambda op: op in parsing_from,
            rp_expr=lambda f, t: f in self._precomputed_op_fanin[t.name],
            ep_expr=None,
            direction="down",
        )
        return fanout

    def parse_transitive_fanin(self, parsing_from: List[Operation]) -> Set[Operation]:
        """从指定节点出发，寻找所有扇出节点
        注意此函数与 graph.get_upstream_operations 的区别
        graph.get_upstream_operations 函数将直接返回目标节点的所有上游节点
        而此函数将只返回上游节点中与当前节点存在数值传递关系的节点

        例如对于子图 conv - shape
            graph.get_upstream_operations(shape) 将返回 conv 节点
            parse_transitive_fanin(shape) 将返回空
        """
        fanin = self._search_engine.opset_matching(
            sp_expr=lambda op: op in parsing_from,
            rp_expr=lambda f, t: t in self._precomputed_op_fanin[f.name],
            ep_expr=None,
            direction="up",
        )
        return fanin

    def dispatch(
        self,
        quant_types: Set[str] = None,
        in_plat: List[TargetPlatform] = None,
        out_plat: List[TargetPlatform] = None,
        **kwargs,
    ) -> Dict[str, TargetPlatform]:
        """对当前图执行默认算子调度逻辑

        在 Onnx 定义的神经网络中存在两类数据：
            其一是从 Gemm, Conv 等计算节点出发的数据，他们往往是需要量化的；
            以及从 Shape, TopK 出发的数据，他们往往是不能被量化的；

        为此 PPQ 将从网络中所有计算节点出发，求解关于计算节点的传递闭包 A
        而后从 Shape, TopK 等节点出发，求解非计算节点的传递闭包 B

        集合 A - B 中的节点将被量化，也被称为量化区节点
        集合 A * B 中的节点将被称为冲突区节点，默认不量化，调度到端侧执行
        集合 B 中的节点将被称为 SOI 节点，不量化且调度到 Cpu 执行
        集合 A, B 之外的节点为未知区域节点，默认不量化，调度到端侧执行

        PPQ 使用朴素递归求解传递闭包，对于神经网络这种稀疏图而言，

        其时间复杂度大约为O(n)
        """
        computing_ops = self.mark_quantable_op()
        soi_ops = self.mark_soi_op()

        dispatching_table = {}
        for op in self.graph.operations.values():
            dispatching_table[op.name] = TargetPlatform.FP32

        for op in self.graph.operations.values():
            if op in computing_ops and op in soi_ops:
                dispatching_table[op.name] = TargetPlatform.FP32
            elif op in computing_ops:
                dispatching_table[op.name] = TargetPlatform.UNSPECIFIED
            elif op in soi_ops:
                dispatching_table[op.name] = TargetPlatform.SOI

        return dispatching_table

    def opsocket(self, op: Operation) -> OpSocket:
        return self.sockets[op.name]



import xslim

# common
# xslim.quantize_onnx_model("resnet18.json")

# xslim.quantize_onnx_model("resnet18.json", "/home/share/modelzoo/classification/resnet18/resnet18.onnx")

# xslim.quantize_onnx_model(
#    "resnet18.json", "/home/share/modelzoo/classification/resnet18/resnet18.onnx", "resnet18_output.onnx"
# )

# import onnx
# onnx_model = onnx.load("/home/share/modelzoo/classification/resnet18/resnet18.onnx")
# quantized_onnx_model = xslim.quantize_onnx_model("resnet18.json", onnx_model)

# custom quantize config
# xslim.quantize_onnx_model("mobilenet_v3_small.json")
#
## custom preprocess
# xslim.quantize_onnx_model("resnet18_custom_preprocess.json")
#
## bert
# xslim.quantize_onnx_model("bertsquad.json")

## dynamic quantize
# xslim.quantize_onnx_model("mobilenet_v3_small_dyn_quantize.json")



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
from typing import Callable, Dict, Iterable, List, Sequence, Set, Tuple, Union

import torch
from xslim.logger import logger

from ..defs import BIAS_CORRECTION_INTERST_TYPE, PASSIVE_OPERATIONS, XQUANT_CONFIG
from ..ppq_decorator import (
    BaseGraph,
    BaseGraphExecutor,
    Operation,
    QuantableOperation,
    QuantizationOptimizationPass,
    QuantizationStates,
    SearchableGraph,
    TargetPlatform,
    Variable,
)


class FlattenGemmFusionPass(QuantizationOptimizationPass):
    def __init__(self) -> None:
        super().__init__("XSlim FlattenGemmFusionPass")

    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Iterable,
        executor: BaseGraphExecutor,
        **kwargs,
    ) -> None:
        search_engine = SearchableGraph(graph)
        paths = search_engine.path_matching(
            sp_expr=lambda x: x.type in {"Flatten"} and x.attributes.get("axis", 0) == 1,
            rp_expr=lambda x, y: False,
            ep_expr=lambda x: x.type in {"Gemm"}
            and x.attributes.get("alpha", 1) == 1
            and x.attributes.get("transA", 0) == 0
            and len(x.inputs) >= 2
            and x.inputs[1].is_parameter,
            direction="down",
        )

        for path in paths:
            path = path.tolist()
            assert len(path) == 2, "Oops seems we got something unexpected."

            flatten_op, gemm_op = path
            assert isinstance(flatten_op, Operation) and isinstance(gemm_op, Operation)

            transB = gemm_op.attributes.get("transB", 0)

            w = gemm_op.parameters[0].value
            if transB != 1:
                w = torch.permute(w, [1, 0])

            if not isinstance(flatten_op.inputs[0].shape, Sequence):
                continue

            weight_new_shape = [*w.shape, *flatten_op.inputs[0].shape[2:]]
            weight_new_shape[1] = -1
            w = w.reshape(weight_new_shape)
            gemm_op.inputs[1].value = w
            conv_attributes = {"dilations": [1, 1], "kernel_shape": weight_new_shape[2:], "strides": [1, 1], "group": 1}
            gemm_op.type = "Conv"
            gemm_op.attributes.clear()
            for k, v in conv_attributes.items():
                gemm_op.attributes[k] = v

            gemm_op.inputs[0] = flatten_op.inputs[0]
            gemm_op.inputs[0].dest_ops.remove(flatten_op)
            gemm_op.inputs[0].dest_ops.append(gemm_op)

            temp_var = flatten_op.outputs[0]
            flatten_op.outputs[0] = gemm_op.outputs[0]
            gemm_op.outputs[0] = temp_var
            flatten_op.inputs[0] = gemm_op.outputs[0]

            flatten_op.inputs[0].dest_ops.remove(gemm_op)
            flatten_op.inputs[0].dest_ops.append(flatten_op)

            gemm_op.outputs[0].source_op = gemm_op
            flatten_op.outputs[0].source_op = flatten_op
            flatten_op.platform = TargetPlatform.UNSPECIFIED
            gemm_op.platform = TargetPlatform.UNSPECIFIED

        for op_name, op in graph.operations.items():
            if op.type in {"Gemm"}:
                if op.inputs[1].is_parameter and op.attributes.get("transB", 0) == 0:
                    op.attributes["transB"] = 1
                    op.inputs[1].value = torch.permute(op.inputs[1].value, [1, 0])


class FormatBatchNormalizationPass(QuantizationOptimizationPass):
    def __init__(self) -> None:
        super().__init__("XSlim FormatBatchNormalizationPass")

    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Iterable,
        executor: BaseGraphExecutor,
        **kwargs,
    ) -> None:
        op_list = [op for op in graph.operations.values()]
        for op in op_list:
            if op.type == "BatchNormalization" and op.inputs[1].is_parameter and op.inputs[2].is_parameter:
                if not isinstance(op.inputs[0].shape, Sequence):
                    raise RuntimeError("remove BatchNormalization error, shape not found.")
                weight_shape = [op.inputs[0].shape[1]] + [1] * (len(op.inputs[0].shape) - 2)
                alpha = op.parameters[0].value
                beta = op.parameters[1].value
                mean = op.parameters[2].value
                var = op.parameters[3].value
                epsilon = op.attributes.get("epsilon", 1e-5)

                with torch.no_grad():
                    w = alpha / torch.sqrt(var + epsilon)
                    w = w.reshape(weight_shape)
                    b = alpha * (-mean) / torch.sqrt(var + epsilon) + beta
                    b = b.reshape(weight_shape)

                op.type = "Mul"
                op.attributes.clear()

                graph.remove_variable(op.inputs[-1])
                graph.remove_variable(op.inputs[-1])
                bias_var = op.inputs.pop()
                bias_var.dest_ops.remove(op)
                with torch.no_grad():
                    op.inputs[1].value = w
                    bias_var.value = b
                op.platform = TargetPlatform.UNSPECIFIED
                bias_out_var = op.outputs[0]
                inner_var = Variable("{}_bn_inner".format(op.name), shape=op.inputs[0].shape)
                graph.append_variable(inner_var)
                op.outputs[0] = inner_var
                inner_var.source_op = op

                bias_op = Operation(
                    "{}_bias".format(op.name),
                    "Add",
                    attributes={},
                    inputs=[inner_var, bias_var],
                    outputs=[bias_out_var],
                    platform=TargetPlatform.UNSPECIFIED,
                )
                inner_var.dest_ops.append(bias_op)
                bias_var.dest_ops.append(bias_op)
                bias_out_var.source_op = bias_op
                graph.append_operation(bias_op)


class HardSwishFusionPass(QuantizationOptimizationPass):
    def __init__(self) -> None:
        super().__init__("XSlim HardSwish Fusion")

    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Iterable,
        executor: BaseGraphExecutor,
        **kwargs,
    ) -> None:
        search_engine = SearchableGraph(graph)
        patterns = search_engine.pattern_matching(
            patterns=[lambda x: x.is_computing_op, "HardSigmoid", "Mul"],
            edges=[[0, 1], [1, 2], [0, 2]],
            exclusive=True,
        )

        for pattern in patterns:
            if any([not isinstance(op, QuantableOperation) for op in pattern]):
                logger.warning(
                    f"There is a pattern of swish activation in your network start from {pattern[0]}, "
                    "however part of your swish activation is not quantable, "
                    "so that graph fusion can not merge their quantization configuration."
                )
                continue
            if any([op.platform != pattern[0].platform for op in pattern]):
                logger.warning(
                    f"There is a pattern of swish activation in your network start from {pattern[0]}, "
                    "however part of your swish activation is not quantable, "
                    "so that graph fusion can not merge their quantization configuration."
                )
                continue
            computing, sigmoid, mul = pattern

            assert isinstance(computing, QuantableOperation)
            assert isinstance(sigmoid, QuantableOperation)
            assert isinstance(mul, QuantableOperation)

            computing_config = computing.config.output_quantization_config[0]
            sigmoid.config.input_quantization_config[0].dominated_by = computing_config
            sigmoid.config.output_quantization_config[0].state = QuantizationStates.FP32
            mul.config.input_quantization_config[0].dominated_by = computing_config
            mul.config.input_quantization_config[1].state = QuantizationStates.FP32


class SwishFusionPass(QuantizationOptimizationPass):
    def __init__(self) -> None:
        super().__init__("XSlim Swish Fusion")

    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Iterable,
        executor: BaseGraphExecutor,
        **kwargs,
    ) -> None:
        search_engine = SearchableGraph(graph)
        patterns = search_engine.pattern_matching(
            patterns=[lambda x: x.is_computing_op, "Sigmoid", "Mul"],
            edges=[[0, 1], [1, 2], [0, 2]],
            exclusive=True,
        )

        for pattern in patterns:
            if any([not isinstance(op, QuantableOperation) for op in pattern]):
                logger.warning(
                    f"There is a pattern of swish activation in your network start from {pattern[0]}, "
                    "however part of your swish activation is not quantable, "
                    "so that graph fusion can not merge their quantization configuration."
                )
                continue
            if any([op.platform != pattern[0].platform for op in pattern]):
                logger.warning(
                    f"There is a pattern of swish activation in your network start from {pattern[0]}, "
                    "however part of your swish activation is not quantable, "
                    "so that graph fusion can not merge their quantization configuration."
                )
                continue
            computing, sigmoid, mul = pattern

            assert isinstance(computing, QuantableOperation)
            assert isinstance(sigmoid, QuantableOperation)
            assert isinstance(mul, QuantableOperation)

            computing_config = computing.config.output_quantization_config[0]
            sigmoid.config.input_quantization_config[0].dominated_by = computing_config
            sigmoid.config.output_quantization_config[0].state = QuantizationStates.FP32
            mul.config.input_quantization_config[0].dominated_by = computing_config
            mul.config.input_quantization_config[1].state = QuantizationStates.FP32


class ComputingFusionPass(QuantizationOptimizationPass):
    def __init__(self, optimize_count: int = 2) -> None:
        super().__init__("XSlim Computing Ops Fusion")
        self.optimize_count = optimize_count

    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Iterable,
        executor: BaseGraphExecutor,
        **kwargs,
    ) -> None:
        search_engine = SearchableGraph(graph=graph)
        paths = search_engine.path_matching(
            sp_expr=lambda x: x.type in {"Conv", "Gemm", "ConvTranspose"} and x.num_of_parameter > 0,
            rp_expr=lambda x, y: False,
            ep_expr=lambda x: x.type in {"Mul", "Div", "Add", "Sub"}
            and any([in_var.is_parameter for in_var in x.inputs]),
            direction="down",
        )

        for path in paths:
            path = path.tolist()
            assert len(path) == 2, "Oops seems we got something unexpected."

            computing_op, mul_op = path
            assert isinstance(computing_op, Operation) and isinstance(mul_op, Operation)

            if (
                len(graph.get_downstream_operations(computing_op)) != 1
                or len(graph.get_upstream_operations(mul_op)) != 1
            ):
                logger.warning(
                    f"PPQ can not merge operation {computing_op.name} and {mul_op.name}, "
                    "this is not suppose to happen with your network, "
                    "network with batchnorm inside might not be able to quantize and deploy."
                )
                continue

            parameter_index = [in_var.is_parameter for in_var in mul_op.inputs].index(True)
            feature_index = 1 - parameter_index
            parameter = mul_op.inputs[parameter_index]
            feature = mul_op.inputs[feature_index]

            if computing_op.num_of_parameter == 1:
                w = computing_op.parameters[0].value  # no bias.
                assert isinstance(w, torch.Tensor), "values of parameters are assumed as torch Tensor"
                if computing_op.type == "ConvTranspose":
                    b = torch.zeros(w.shape[1] * computing_op.attributes.get("group", 1)).to(w.device)
                elif computing_op.type == "Gemm" and computing_op.attributes.get("transB", 0) == 0:
                    b = torch.zeros(w.shape[1]).to(w.device)
                else:
                    b = torch.zeros(w.shape[0]).to(w.device)
            else:
                w, b = [var.value for var in computing_op.parameters[:2]]  # has bias.

            if parameter.value.numel() == 1 or parameter.value.ndim == w.ndim:
                pass
            else:
                continue

            if parameter.value.numel() == 1 or (
                parameter.value.ndim >= 2 and parameter.value.shape[1] == parameter.value.numel()
            ):
                pass
            else:
                continue

            if mul_op.type in {"Mul", "Div"}:
                alpha = parameter.value

                if mul_op.type == "Div":
                    if parameter_index != 1:
                        continue
                    alpha = 1.0 / alpha

                if computing_op.type == "Conv":
                    # calculate new weight and bias
                    scale = alpha
                    w = w * scale.reshape([-1] + [1] * (w.ndim - 1))
                    b = alpha * b

                elif computing_op.type == "Gemm":
                    # calculate new weight and bias
                    scale = alpha
                    if computing_op.attributes.get("transB", 0):
                        w = w * scale.reshape([-1, 1])
                    else:
                        w = w * scale.reshape([1, -1])
                    b = alpha * b

                elif computing_op.type == "ConvTranspose":
                    scale = alpha
                    group = computing_op.attributes.get("group", 1)
                    scale = scale.reshape([group, 1, -1] + [1] * (w.ndim - 2))
                    w = w.reshape([group, -1] + list(w.shape[1:])) * scale
                    w = w.reshape([w.shape[0] * w.shape[1]] + list(w.shape[2:]))
                    b = alpha * b
                else:
                    raise TypeError(
                        f"Unexpected op type {computing_op.type}. "
                        f"Can not merge {computing_op.name} with {mul_op.name}"
                    )
            else:
                beta = parameter.value
                if mul_op.type == "Div":
                    if parameter_index != 1:
                        continue
                    beta = -beta
                b = beta.reshape(-1) + b

            # create new op and variable
            merged_op = Operation(
                computing_op.name, op_type=computing_op.type, attributes=computing_op.attributes.copy()
            )
            weight_var = Variable(computing_op.name + "_weight", w, True, [merged_op])
            bias_var = Variable(computing_op.name + "_bias", b, True, [merged_op])

            # replace & dirty work
            input_var = computing_op.inputs[0]
            output_var = mul_op.outputs[0]

            input_var.dest_ops.remove(computing_op)
            input_var.dest_ops.append(merged_op)

            output_var.source_op = merged_op

            # delete old operations
            computing_op.inputs.pop(0)
            mul_op.outputs.clear()
            graph.remove_operation(computing_op)
            graph.remove_operation(mul_op)

            # insert new
            graph.append_operation(merged_op)
            merged_op.inputs.extend([input_var, weight_var, bias_var])
            merged_op.outputs.extend([output_var])

            graph.append_variable(weight_var)
            graph.append_variable(bias_var)

        self.optimize_count -= 1
        if self.optimize_count > 0:
            self.optimize(graph, dataloader, executor, **kwargs)



from typing import Callable, List, Union

import torch

from ..core import DataType

from .base.graph import BaseGraph
from .processer import GraphCommandProcessor


class TrainableGraph(GraphCommandProcessor):
    """Trainable Graph offers a bunch of functions that provide training interfaces."""

    def __init__(self, graph_or_processor: Union[BaseGraph, Callable]) -> None:
        super().__init__(graph_or_processor)

    def parameters(self) -> List[torch.Tensor]:
        parameters = []
        for var in self.graph.variables.values():
            if var.is_parameter and DataType.to_torch(var.dtype) == torch.float:
                parameters.append(var.value)
        return parameters

    def zero_grad(self):
        for var in self.graph.variables.values():
            if var.is_parameter and DataType.to_torch(var.dtype) == torch.float:
                if var.value._grad is not None:
                    var.value._grad.zero_()

    def state_dict(self) -> dict:
        parameters = {}
        for var in self.graph.variables.values():
            if var.is_parameter and DataType.to_torch(var.dtype) == torch.float:
                parameters[var.name] = var.value
        return parameters

    def _acceptable_command_types(self):
        return None

    def process(self):
        return None



from typing import Callable, Dict, Iterator, List

import torch
from ...core import PASSIVE_OPERATIONS, ppq_warning
from ...executor import RuntimeHook, TorchExecutor
from ...IR import BaseGraph, Operation, QuantableOperation, Variable
from ...quantization.measure.norm import torch_snr_error
from ...utils.fetch import batch_random_fetch, tensor_random_fetch
from tqdm import tqdm

from .util import MeasurePrinter, MeasureRecorder


class OutputRecorder(RuntimeHook):
    def __init__(self, operation: Operation, fetchs: int = 4096) -> None:
        self.fetched = None
        self.fetchs = fetchs
        super().__init__(operation)

    def pre_forward_hook(self, inputs: list, **kwargs) -> list:
        return super().pre_forward_hook(inputs, **kwargs)

    def post_forward_hook(self, outputs: list, **kwargs) -> list:
        output_tensor = outputs[0]
        assert isinstance(output_tensor, torch.Tensor), "Output of monitoring operation is not a torch.Tensor"
        self.fetched = batch_random_fetch(output_tensor, seed=10086, fetches_per_batch=self.fetchs).to("cpu")
        return super().post_forward_hook(outputs, **kwargs)

    def pop(self) -> torch.Tensor:
        fetched = self.fetched
        self.fetched = None
        return fetched


class DetailedRecorder(RuntimeHook):
    def __init__(self, operation: Operation, fetchs: int = 1024) -> None:
        self.fetchs = fetchs
        self.i_storage = [[] for _ in range(operation.num_of_input)]
        self.o_storage = [[] for _ in range(operation.num_of_output)]
        super().__init__(operation)

    def pre_forward_hook(self, inputs: List[torch.Tensor], **kwargs) -> list:
        for idx, input in enumerate(inputs):
            self.i_storage[idx].append(tensor_random_fetch(input, seed=10086, num_of_fetches=self.fetchs).to("cpu"))
        return super().pre_forward_hook(inputs, **kwargs)

    def post_forward_hook(self, outputs: List[torch.Tensor], **kwargs) -> list:
        for idx, output in enumerate(outputs):
            self.o_storage[idx].append(tensor_random_fetch(output, seed=10086, num_of_fetches=self.fetchs).to("cpu"))
        return super().post_forward_hook(outputs, **kwargs)

    def clear(self):
        self.i_storage = [[] for _ in range(self._hook_to.num_of_input)]
        self.o_storage = [[] for _ in range(self._hook_to.num_of_output)]


def graphwise_error_analyse(
    graph: BaseGraph,
    running_device: str,
    dataloader: Iterator,
    collate_fn: Callable = None,
    method: str = "snr",
    steps: int = 8,
    verbose: bool = True,
    fetchs: int = 4096,
) -> Dict[str, float]:
    """Measure the difference from a quantized graph to its dequantized graph.

    A dictionary contains output differences for all operation will be returned as a result.

        Result is like: {'operation name 1': 0.933, 'operation name 2': 0.926}

    if verbose is set as True, this function will display error report at last.

    The key of the dictionary is an operation name while the value of corresponding key
        is the difference between quantized output and float output of this operation.

    Result {'operation name 1': 0.933} means that quantized graph and fp32 graph have a difference
        (or similarity, based on your measurement) 0.933 at the output tensor of 'operation name 1'.

    ATTENTION: Output difference is measured at graph-level, it includes the difference accmulated from the
        very beginning operation along to the target operation.

    Args:
        graph (BaseGraph):
            A fully quantized graph instance.

        running_device (str):
            A device string used to initialize a graph executor for the graph execution.
                if a executor was given, this parameter will be skipped.

        dataloader (Iterator):
            Test dataloader, this function will measure the output difference based on given data.

        collate_fn (Callable, optional):
            An data preprocessing function provided by user to convert data from dataloader towards
                executable format. If set as None, then no action will be taken during preprocessing.

        method (str, optional):
            A string indicates a measurement to calculate the difference of quantized output and fp32 one.
                'cosine', 'snr', and 'mse' is supported in PPQ for now.

        steps (Int, optional)
            computation steps.

    Returns:
        A dictionary contains output differences for all operation will be returned from this function.

        Result is like: {'operation name 1': 0.933, 'operation name 2': 0.926}
    """
    executor = TorchExecutor(graph=graph, device=running_device)

    # find all quantable operations.
    interested_op = [
        operation
        for operation in graph.operations.values()
        if (isinstance(operation, QuantableOperation) and operation.is_computing_op)
    ]
    if len(interested_op) == 0:
        print("Oops. you got nothing to analyse.")
        return

    # set up all hooks.
    recorders, hooks, caches = {}, {}, {}
    for operation in interested_op:
        if isinstance(operation, QuantableOperation):
            if operation.num_of_output > 1:
                ppq_warning(
                    f"Operation {operation.name} has more than 1 output, "
                    "analyser will process the first output of it."
                )

            recorders[operation.name] = MeasureRecorder(measurement=method)
            hooks[operation.name] = OutputRecorder(operation=operation, fetchs=fetchs)
            caches[operation.name] = []

    # dequantize all
    for operation in graph.operations.values():
        if isinstance(operation, QuantableOperation):
            operation.dequantize()

    # run for each quantable operations:
    for idx, batch in tqdm(
        enumerate(dataloader),
        desc="Analysing Graphwise Quantization Error(Phrase 1):",
        total=(min(len(dataloader), steps)),
    ):
        if collate_fn is not None:
            batch = collate_fn(batch)
        executor.forward(inputs=batch, hooks=hooks)

        for operation in interested_op:
            hook = hooks[operation.name]
            caches[operation.name].append(hook.pop())

        if idx >= steps:
            break

    # restore all
    for operation in graph.operations.values():
        if isinstance(operation, QuantableOperation):
            operation.restore_quantize_state()

    # run for each quantable operations:
    for idx, batch in tqdm(
        enumerate(dataloader),
        desc="Analysing Graphwise Quantization Error(Phrase 2):",
        total=(min(len(dataloader), steps)),
    ):
        if collate_fn is not None:
            batch = collate_fn(batch)
        executor.forward(inputs=batch, hooks=hooks)

        for operation in interested_op:
            recorder = recorders[operation.name]
            hook = hooks[operation.name]
            cache = caches[operation.name]
            recorder.update(y_real=cache[idx], y_pred=hook.pop())

        if idx >= steps:
            break

    results = {}
    for operation in interested_op:
        assert isinstance(operation, QuantableOperation)
        results[operation.name] = recorders[operation.name].measure

    if verbose:
        method_str = "MEASUREMENT"
        if method == "snr":
            method_str = "NOISE:SIGNAL POWER RATIO"
        if method == "cosine":
            method_str = "COSINE SIMILARITY"
        if method == "mse":
            method_str = "MSE LOSS(UNSCALED)"
        MeasurePrinter(
            results, order="large_to_small", measure=method_str, percentage=method in {"snr", "cosine"}
        ).print()
    return results


def statistical_analyse(
    graph: BaseGraph,
    running_device: str,
    dataloader: Iterator,
    collate_fn: Callable = None,
    steps: int = 8,
) -> List[dict]:
    """It is time to do some statistical work.

    statistical_analyse is a powerful analying function
        that provides a in-depth study of your network.

    use report = statistical_analyse() to invoke this function

    The return value of this function is a collection of statistics parameters
    You are recommended to processing them with pandas

    from pandas import DataFrame
    report_df = DataFrame(report)

    Args:
        graph (BaseGraph): _description_
        running_device (str): _description_
        dataloader (Iterator): _description_
        collate_fn (Callable, optional): _description_. Defaults to None.
        steps (int, optional): _description_. Defaults to 8.

    Returns:
        Dict[str, float]: _description_
    """

    class StatisticalErrorAnalyser:
        def __init__(self, x_fp: List[torch.Tensor], x_qt: List[torch.Tensor], op: Operation, var: Variable) -> None:
            self.x_qt = torch.cat(x_qt, dim=0)
            self.x_fp = torch.cat(x_fp, dim=0)
            self.x_er = self.x_qt - self.x_fp
            self.op = op
            self.var = var

            self.num_of_samples = self.x_fp.shape[0]

        def stat(self) -> dict:
            x_er, x_fp, x_qt = self.x_er, self.x_fp, self.x_qt
            er_mean = x_er.mean().item()
            er_std = x_er.std().item()
            er_min = x_er.min().item()
            er_max = x_er.max().item()
            er_skew = self.solve_skewness(x_er, er_mean, er_std).item()
            er_kurtosis = self.solve_kurtosis(x_er, er_mean, er_std).item()
            er_hist = torch.histc(x_er, bins=32, min=x_er.min(), max=x_er.max()).cpu().tolist()

            qt_mean = x_qt.mean().item()
            qt_std = x_qt.std().item()
            qt_min = x_qt.min().item()
            qt_max = x_qt.max().item()
            qt_skew = self.solve_skewness(x_qt, qt_mean, qt_std).item()
            qt_kurtosis = self.solve_kurtosis(x_qt, qt_mean, qt_std).item()
            qt_hist = torch.histc(x_qt, bins=32, min=x_qt.min(), max=x_qt.max()).cpu().tolist()

            fp_mean = x_fp.mean().item()
            fp_std = x_fp.std().item()
            fp_min = x_fp.min().item()
            fp_max = x_fp.max().item()
            fp_skew = self.solve_skewness(x_fp, fp_mean, fp_std).item()
            fp_kurtosis = self.solve_kurtosis(x_fp, fp_mean, fp_std).item()
            fp_hist = torch.histc(x_fp, bins=32, min=x_fp.min(), max=x_fp.max()).cpu().tolist()

            snr = torch_snr_error(x_qt, x_fp).item()
            return {
                "Op name": self.op.name,
                "Op type": self.op.type,
                "Is parameter": self.var.is_parameter,
                "Is input": self.var in self.op.inputs,
                "Is output": self.var in self.op.outputs,
                "Variable name": self.var.name,
                "Noise:Signal Power Ratio": snr,
                "Noise Mean": er_mean,
                "Noise Std": er_std,
                "Noise Skewness": er_skew,
                "Noise Kurtosis": er_kurtosis,
                "Noise Hist": er_hist,
                "Noise Max": er_max,
                "Noise Min": er_min,
                "Quantized Mean": qt_mean,
                "Quantized Std": qt_std,
                "Quantized Skewness": qt_skew,
                "Quantized Kurtosis": qt_kurtosis,
                "Quantized Hist": qt_hist,
                "Quantized Max": qt_max,
                "Quantized Min": qt_min,
                "Float Mean": fp_mean,
                "Float Std": fp_std,
                "Float Skewness": fp_skew,
                "Float Kurtosis": fp_kurtosis,
                "Float Hist": fp_hist,
                "Float Max": fp_max,
                "Float Min": fp_min,
            }

        def solve_skewness(self, x: torch.Tensor, mean: float, std: float) -> torch.Tensor:
            # 三次方可能有数据溢出
            return torch.pow((x - mean) / std, 3).mean()

        def solve_kurtosis(self, x: torch.Tensor, mean: float, std: float) -> torch.Tensor:
            # 四次方可能有数据溢出
            return torch.pow((x - mean) / std, 4).mean() - 3

    executor = TorchExecutor(graph=graph, device=running_device)
    # find all quantable operations.
    interested_op = []
    for operation in graph.operations.values():
        if isinstance(operation, QuantableOperation) and operation.type not in PASSIVE_OPERATIONS:
            interested_op.append(operation)
    if len(interested_op) == 0:
        print("Oops. you got nothing to analyse.")
        return

    # set up all hooks.
    hooks, caches = {}, {}
    for operation in interested_op:
        if isinstance(operation, QuantableOperation):
            hooks[operation.name] = DetailedRecorder(operation=operation)
            caches[operation.name] = {
                "Quantized Input": [],
                "Quantized Output": [],
                "Dequantized Input": [],
                "Dequantized Output": [],
            }

    # dequantize all
    for operation in graph.operations.values():
        if isinstance(operation, QuantableOperation):
            operation.dequantize()

    # run for each quantable operations:
    for idx, batch in tqdm(enumerate(dataloader), desc="Analysing Phrase 1", total=(min(len(dataloader), steps))):
        if collate_fn is not None:
            batch = collate_fn(batch)
        executor.forward(inputs=batch, hooks=hooks)
        if idx >= steps:
            break

    for operation in interested_op:
        hook = hooks[operation.name]
        assert isinstance(hook, DetailedRecorder)
        caches[operation.name]["Dequantized Input"] = hook.i_storage.copy()
        caches[operation.name]["Dequantized Output"] = hook.o_storage.copy()
        hook.clear()

    # restore all
    for operation in graph.operations.values():
        if isinstance(operation, QuantableOperation):
            operation.restore_quantize_state()

    # run for each quantable operations:
    for idx, batch in tqdm(enumerate(dataloader), desc="Analysing Phrase 2", total=(min(len(dataloader), steps))):
        if collate_fn is not None:
            batch = collate_fn(batch)
        executor.forward(inputs=batch, hooks=hooks)
        if idx >= steps:
            break

    for operation in interested_op:
        hook = hooks[operation.name]
        assert isinstance(hook, DetailedRecorder)
        caches[operation.name]["Quantized Input"] = hook.i_storage.copy()
        caches[operation.name]["Quantized Output"] = hook.o_storage.copy()
        hook.clear()

    # analysing cache
    records = []
    for name, record in caches.items():
        operation = graph.operations[name]
        assert isinstance(operation, Operation)
        for idx, input_var in enumerate(operation.inputs):
            x_qt = record["Quantized Input"][idx]
            x_fp = record["Dequantized Input"][idx]
            records.append(StatisticalErrorAnalyser(x_fp=x_fp, x_qt=x_qt, op=operation, var=input_var).stat())

        for idx, output_var in enumerate(operation.outputs):
            x_qt = record["Quantized Output"][idx]
            x_fp = record["Dequantized Output"][idx]
            records.append(StatisticalErrorAnalyser(x_fp=x_fp, x_qt=x_qt, op=operation, var=output_var).stat())

    return records
